# Copyright 2013 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Management class for migration / resize operations.
"""
import os
import re

from nova import block_device
import nova.conf
from nova import exception
from nova.virt import configdrive
from nova.virt import driver
from os_win import utilsfactory
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import units

from compute_hyperv.i18n import _
from compute_hyperv.nova import block_device_manager
from compute_hyperv.nova import constants
from compute_hyperv.nova import imagecache
from compute_hyperv.nova import pathutils
from compute_hyperv.nova import vmops
from compute_hyperv.nova import volumeops

LOG = logging.getLogger(__name__)
CONF = nova.conf.CONF


class MigrationOps(object):

    _ADMINISTRATIVE_SHARE_RE = re.compile(r'\\\\.*\\[a-zA-Z]\$\\.*')

    def __init__(self):
        self._vmutils = utilsfactory.get_vmutils()
        self._vhdutils = utilsfactory.get_vhdutils()
        self._pathutils = pathutils.PathUtils()
        self._volumeops = volumeops.VolumeOps()
        self._vmops = vmops.VMOps()
        self._imagecache = imagecache.ImageCache()
        self._block_dev_man = block_device_manager.BlockDeviceInfoManager()
        self._migrationutils = utilsfactory.get_migrationutils()
        self._metricsutils = utilsfactory.get_metricsutils()

    def _move_vm_files(self, instance):
        instance_path = self._pathutils.get_instance_dir(instance.name)
        revert_path = self._pathutils.get_instance_migr_revert_dir(
            instance_path, remove_dir=True, create_dir=True)
        export_path = self._pathutils.get_export_dir(
            instance_dir=revert_path, create_dir=True)

        # copy the given instance's files to a _revert folder, as backup.
        LOG.debug("Moving instance files to a revert path: %s",
                  revert_path, instance=instance)
        self._pathutils.move_folder_files(instance_path, revert_path)
        self._pathutils.copy_vm_config_files(instance.name, export_path)

        return revert_path

    def _check_target_flavor(self, instance, flavor, block_device_info):
        ephemerals = driver.block_device_info_get_ephemerals(block_device_info)
        eph_size = (block_device.get_bdm_ephemeral_disk_size(ephemerals) or
                    instance.flavor.ephemeral_gb)

        new_root_gb = flavor.root_gb
        curr_root_gb = instance.flavor.root_gb
        new_eph_size = flavor.ephemeral_gb

        root_down = new_root_gb < curr_root_gb
        ephemeral_down = new_eph_size < eph_size
        booted_from_volume = self._block_dev_man.is_boot_from_volume(
            block_device_info)

        if root_down and not booted_from_volume:
            raise exception.InstanceFaultRollback(
                exception.CannotResizeDisk(
                    reason=_("Cannot resize the root disk to a smaller size. "
                             "Current size: %(curr_root_gb)s GB. Requested "
                             "size: %(new_root_gb)s GB.") % {
                                 'curr_root_gb': curr_root_gb,
                                 'new_root_gb': new_root_gb}))
        # We allow having a new flavor with no ephemeral storage, in which
        # case we'll just remove all the ephemeral disks.
        elif ephemeral_down and new_eph_size:
            reason = (_("The new flavor ephemeral size (%(flavor_eph)s) is "
                       "smaller than the current total ephemeral disk size: "
                       "%(current_eph)s.") %
                      dict(flavor_eph=flavor.ephemeral_gb,
                           current_eph=eph_size))
            raise exception.InstanceFaultRollback(
                exception.CannotResizeDisk(reason=reason))

    def migrate_disk_and_power_off(self, context, instance, dest,
                                   flavor, network_info,
                                   block_device_info=None, timeout=0,
                                   retry_interval=0):
        LOG.debug("migrate_disk_and_power_off called", instance=instance)

        self._check_target_flavor(instance, flavor, block_device_info)

        self._vmops.power_off(instance, timeout, retry_interval)
        instance_path = self._move_vm_files(instance)

        instance.system_metadata['backup_location'] = instance_path
        instance.save()

        self._vmops.destroy(instance, network_info,
                            block_device_info, destroy_disks=True,
                            cleanup_migration_files=False)

        # return the instance's path location.
        return instance_path

    def confirm_migration(self, context, migration, instance, network_info):
        LOG.debug("confirm_migration called", instance=instance)
        revert_path = instance.system_metadata['backup_location']
        export_path = self._pathutils.get_export_dir(instance_dir=revert_path)
        self._pathutils.check_dir(export_path, remove_dir=True)
        self._pathutils.check_dir(revert_path, remove_dir=True)

    def _revert_migration_files(self, instance):
        revert_path = instance.system_metadata['backup_location']
        instance_path = re.sub('_revert$', '', revert_path)

        # the instance dir might still exist, if the destination node kept
        # the files on the original node.
        self._pathutils.check_dir(instance_path, remove_dir=True)
        self._pathutils.rename(revert_path, instance_path)
        return instance_path

    def _check_and_attach_config_drive(self, instance, vm_gen):
        if configdrive.required_by(instance):
            configdrive_path = self._pathutils.lookup_configdrive_path(
                instance.name)
            if configdrive_path:
                self._vmops.attach_config_drive(instance, configdrive_path,
                                                vm_gen)
            else:
                raise exception.ConfigDriveNotFound(
                    instance_uuid=instance.uuid)

    def finish_revert_migration(self, context, instance, network_info,
                                block_device_info=None, power_on=True):
        LOG.debug("finish_revert_migration called", instance=instance)
        instance_path = self._revert_migration_files(instance)

        image_meta = self._imagecache.get_image_details(context, instance)
        self._import_and_setup_vm(context, instance, instance_path, image_meta,
                                  block_device_info)

        if power_on:
            self._vmops.power_on(instance, network_info=network_info)

    def _merge_base_vhd(self, diff_vhd_path, base_vhd_path):
        base_vhd_copy_path = os.path.join(os.path.dirname(diff_vhd_path),
                                          os.path.basename(base_vhd_path))
        try:
            LOG.debug('Copying base disk %(base_vhd_path)s to '
                      '%(base_vhd_copy_path)s',
                      {'base_vhd_path': base_vhd_path,
                       'base_vhd_copy_path': base_vhd_copy_path})
            self._pathutils.copyfile(base_vhd_path, base_vhd_copy_path)

            LOG.debug("Reconnecting copied base VHD "
                      "%(base_vhd_copy_path)s and diff "
                      "VHD %(diff_vhd_path)s",
                      {'base_vhd_copy_path': base_vhd_copy_path,
                       'diff_vhd_path': diff_vhd_path})
            self._vhdutils.reconnect_parent_vhd(diff_vhd_path,
                                                base_vhd_copy_path)

            LOG.debug("Merging differential disk %s into its parent.",
                      diff_vhd_path)
            self._vhdutils.merge_vhd(diff_vhd_path)

            # Replace the differential VHD with the merged one
            self._pathutils.rename(base_vhd_copy_path, diff_vhd_path)
        except Exception:
            with excutils.save_and_reraise_exception():
                if self._pathutils.exists(base_vhd_copy_path):
                    self._pathutils.remove(base_vhd_copy_path)

    def _check_resize_vhd(self, vhd_path, vhd_info, new_size):
        curr_size = vhd_info['VirtualSize']
        if new_size < curr_size:
            raise exception.CannotResizeDisk(
                reason=_("Cannot resize the root disk to a smaller size. "
                         "Current size: %(curr_root_gb)s GB. Requested "
                         "size: %(new_root_gb)s GB.") % {
                             'curr_root_gb': curr_size / units.Gi,
                             'new_root_gb': new_size / units.Gi})
        elif new_size > curr_size:
            self._resize_vhd(vhd_path, new_size)

    def _resize_vhd(self, vhd_path, new_size):
        if vhd_path.split('.')[-1].lower() == "vhd":
            LOG.debug("Getting parent disk info for disk: %s", vhd_path)
            base_disk_path = self._vhdutils.get_vhd_parent_path(vhd_path)
            if base_disk_path:
                # A differential VHD cannot be resized. This limitation
                # does not apply to the VHDX format.
                self._merge_base_vhd(vhd_path, base_disk_path)
        LOG.debug("Resizing disk \"%(vhd_path)s\" to new max "
                  "size %(new_size)s",
                  {'vhd_path': vhd_path, 'new_size': new_size})
        self._vhdutils.resize_vhd(vhd_path, new_size)

    def _check_base_disk(self, context, instance, diff_vhd_path,
                         src_base_disk_path):
        base_vhd_path = self._imagecache.get_cached_image(context, instance)

        # If the location of the base host differs between source
        # and target hosts we need to reconnect the base disk
        if src_base_disk_path.lower() != base_vhd_path.lower():
            LOG.debug("Reconnecting copied base VHD "
                      "%(base_vhd_path)s and diff "
                      "VHD %(diff_vhd_path)s",
                      {'base_vhd_path': base_vhd_path,
                       'diff_vhd_path': diff_vhd_path})
            self._vhdutils.reconnect_parent_vhd(diff_vhd_path,
                                                base_vhd_path)

    def _migrate_disks_from_source(self, migration, instance,
                                   source_inst_dir):
        source_inst_dir = self._pathutils.get_remote_path(
            migration.source_compute, source_inst_dir)
        source_export_path = self._pathutils.get_export_dir(
            instance_dir=source_inst_dir)

        if CONF.hyperv.move_disks_on_cold_migration:
            # copy the files from the source node to this node's configured
            # location.
            inst_dir = self._pathutils.get_instance_dir(
                instance.name, create_dir=True, remove_dir=True)
        elif self._ADMINISTRATIVE_SHARE_RE.match(source_inst_dir):
            # make sure that the source is not a remote local path.
            # e.g.: \\win-srv\\C$\OpenStack\Instances\..
            # CSVs, local paths, and shares are fine.
            # NOTE(claudiub): get rid of the final _revert part of the path.
            # rstrip can remove more than _revert, which is not desired.
            inst_dir = re.sub('_revert$', '', source_inst_dir)

            LOG.warning(
                'Host is configured not to copy disks on cold migration, but '
                'the instance will not be able to start with the remote path: '
                '"%s". Only local, share, or CSV paths are acceptable.',
                inst_dir)
            inst_dir = self._pathutils.get_instance_dir(
                instance.name, create_dir=True, remove_dir=True)
        else:
            # make a copy on the source node's configured location.
            # strip the _revert from the source backup dir.
            inst_dir = re.sub('_revert$', '', source_inst_dir)
            self._pathutils.check_dir(inst_dir, create_dir=True)

        export_path = self._pathutils.get_export_dir(
            instance_dir=inst_dir)

        self._pathutils.copy_folder_files(source_inst_dir, inst_dir)
        self._pathutils.copy_dir(source_export_path, export_path)
        return inst_dir

    def finish_migration(self, context, migration, instance, disk_info,
                         network_info, image_meta, resize_instance=False,
                         block_device_info=None, power_on=True):
        LOG.debug("finish_migration called", instance=instance)
        instance_dir = self._migrate_disks_from_source(migration, instance,
                                                       disk_info)

        # NOTE(claudiub): nova compute manager only takes into account disk
        # flavor changes when passing to the driver resize_instance=True.
        # we need to take into account flavor extra_specs as well.
        resize_instance = (
            migration.old_instance_type_id != migration.new_instance_type_id)

        self._import_and_setup_vm(context, instance, instance_dir, image_meta,
                                  block_device_info, resize_instance)

        if power_on:
            self._vmops.power_on(instance, network_info=network_info)

    def _import_and_setup_vm(self, context, instance, instance_dir, image_meta,
                             block_device_info, resize_instance=False):
        vm_gen = self._vmops.get_image_vm_generation(instance.uuid, image_meta)
        self._import_vm(instance_dir)
        self._vmops.update_vm_resources(instance, vm_gen, image_meta,
                                        instance_dir, resize_instance)

        self._volumeops.connect_volumes(block_device_info)
        self._update_disk_image_paths(instance, instance_dir)
        self._check_and_update_disks(context, instance, vm_gen, image_meta,
                                     block_device_info,
                                     resize_instance=resize_instance)
        self._volumeops.fix_instance_volume_disk_paths(
            instance.name, block_device_info)

        self._migrationutils.realize_vm(instance.name)

        # During a resize, ephemeral disks may be removed. We cannot remove
        # disks from a planned vm, for which reason we have to do this after
        # *realizing* it. At the same time, we cannot realize a VM before
        # updating disks to use the destination paths.
        ephemerals = block_device_info['ephemerals']
        self._check_ephemeral_disks(instance, ephemerals, resize_instance)

        self._vmops.configure_remotefx(instance, vm_gen, resize_instance)
        self._vmops.configure_instance_metrics(instance.name)

    def _import_vm(self, instance_dir):
        snapshot_dir = self._pathutils.get_instance_snapshot_dir(
            instance_dir=instance_dir)
        export_dir = self._pathutils.get_export_dir(instance_dir=instance_dir)
        vm_config_file_path = self._pathutils.get_vm_config_file(export_dir)

        self._migrationutils.import_vm_definition(vm_config_file_path,
                                                  snapshot_dir)

        # NOTE(claudiub): after the VM was imported, the VM config files are
        # not necessary anymore.
        self._pathutils.get_export_dir(instance_dir=instance_dir,
                                       remove_dir=True)

    def _update_disk_image_paths(self, instance, instance_path):
        """Checks if disk images have the correct path and updates them if not.

        When resizing an instance, the vm is imported on the destination node
        and the disk files are copied from source node. If the hosts have
        different instance_path config options set, the disks are migrated to
        the correct paths, but vm disk resources are not updated to point to
        the new location.
        """
        (disk_files, volume_drives) = self._vmutils.get_vm_storage_paths(
            instance.name)

        pattern = re.compile('configdrive|eph|root', re.IGNORECASE)
        for disk_file in disk_files:
            disk_name = os.path.basename(disk_file)
            if not pattern.match(disk_name):
                # skip files that do not match the pattern.
                continue

            expected_disk_path = os.path.join(instance_path, disk_name)
            if not os.path.exists(expected_disk_path):
                raise exception.DiskNotFound(location=expected_disk_path)

            if expected_disk_path.lower() != disk_file.lower():
                LOG.debug("Updating VM disk location from %(src)s to %(dest)s",
                          {'src': disk_file, 'dest': expected_disk_path,
                           'instance': instance})
                self._vmutils.update_vm_disk_path(disk_file,
                                                  expected_disk_path,
                                                  is_physical=False)

    def _check_and_update_disks(self, context, instance, vm_gen, image_meta,
                                block_device_info, resize_instance=False):
        self._block_dev_man.validate_and_update_bdi(instance, image_meta,
                                                    vm_gen, block_device_info)
        root_device = block_device_info['root_disk']

        if root_device['type'] == constants.DISK:
            root_vhd_path = self._pathutils.lookup_root_vhd_path(instance.name)
            root_device['path'] = root_vhd_path
            if not root_vhd_path:
                base_vhd_path = self._pathutils.get_instance_dir(instance.name)
                raise exception.DiskNotFound(location=base_vhd_path)

            root_vhd_info = self._vhdutils.get_vhd_info(root_vhd_path)
            src_base_disk_path = root_vhd_info.get("ParentPath")
            if src_base_disk_path:
                self._check_base_disk(context, instance, root_vhd_path,
                                      src_base_disk_path)

            if resize_instance:
                new_size = instance.flavor.root_gb * units.Gi
                self._check_resize_vhd(root_vhd_path, root_vhd_info, new_size)

    def _check_ephemeral_disks(self, instance, ephemerals,
                               resize_instance=False):
        instance_name = instance.name
        new_eph_gb = instance.get('ephemeral_gb', 0)
        ephemerals_to_remove = set()

        if not ephemerals and new_eph_gb:
            # No explicit ephemeral disk bdm was retrieved, yet the flavor
            # provides ephemeral storage, for which reason we're adding a
            # default ephemeral disk.
            eph = dict(device_type='disk',
                       drive_addr=0,
                       size=new_eph_gb)
            ephemerals.append(eph)

        if len(ephemerals) == 1:
            # NOTE(claudiub): Resize only if there is one ephemeral. If there
            # are more than 1, resizing them can be problematic. This behaviour
            # also exists in the libvirt driver and it has to be addressed in
            # the future.
            ephemerals[0]['size'] = new_eph_gb
        elif new_eph_gb and sum(
                eph['size'] for eph in ephemerals) != new_eph_gb:
            # New ephemeral size is different from the original ephemeral size
            # and there are multiple ephemerals.
            LOG.warning("Cannot resize multiple ephemeral disks for instance.",
                        instance=instance)

        for index, eph in enumerate(ephemerals):
            eph_name = "eph%s" % index
            existing_eph_path = self._pathutils.lookup_ephemeral_vhd_path(
                instance_name, eph_name)

            if not existing_eph_path and eph['size']:
                eph['format'] = self._vhdutils.get_best_supported_vhd_format()
                eph['path'] = self._pathutils.get_ephemeral_vhd_path(
                    instance_name, eph['format'], eph_name)
                if not resize_instance:
                    # ephemerals should have existed.
                    raise exception.DiskNotFound(location=eph['path'])

                # We cannot rely on the BlockDeviceInfoManager class to
                # provide us a disk slot as it's only usable when creating
                # new instances (it's not aware of the current disk address
                # layout).
                # There's no way in which IDE may be requested for new
                # ephemeral disks (after a resize), so we'll just enforce
                # SCSI for now. os-win does not currently allow retrieving
                # free IDE slots.
                ctrller_path = self._vmutils.get_vm_scsi_controller(
                    instance.name)
                ctrl_addr = self._vmutils.get_free_controller_slot(
                    ctrller_path)
                eph['disk_bus'] = constants.CTRL_TYPE_SCSI
                eph['ctrl_disk_addr'] = ctrl_addr

                # create ephemerals
                self._vmops.create_ephemeral_disk(instance.name, eph)
                self._vmops.attach_ephemerals(instance_name, [eph])
            elif eph['size'] > 0:
                # ephemerals exist. resize them.
                eph['path'] = existing_eph_path
                eph_vhd_info = self._vhdutils.get_vhd_info(eph['path'])
                self._check_resize_vhd(
                    eph['path'], eph_vhd_info, eph['size'] * units.Gi)
            else:
                eph['path'] = None
                # ephemeral new size is 0, remove it.
                ephemerals_to_remove.add(existing_eph_path)

        if not new_eph_gb:
            # The new flavor does not provide any ephemeral storage. We'll
            # remove any existing ephemeral disk (default ones included).
            attached_ephemerals = self._vmops.get_attached_ephemeral_disks(
                instance.name)
            ephemerals_to_remove |= set(attached_ephemerals)

        for eph_path in ephemerals_to_remove:
            self._vmutils.detach_vm_disk(instance_name, eph_path,
                                         is_physical=False)
            self._pathutils.remove(eph_path)



# Copyright 2015 Intel
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Coordination and locking utilities."""

import inspect
import uuid

import decorator
from nova import exception
from nova import utils
from oslo_config import cfg
from oslo_log import log
from oslo_utils import timeutils
from tooz import coordination

from compute_hyperv.i18n import _

LOG = log.getLogger(__name__)

CONF = cfg.CONF


class Coordinator(object):
    """Tooz coordination wrapper.

    Coordination member id is created from concatenated
    `prefix` and `agent_id` parameters.

    :param str agent_id: Agent identifier
    :param str prefix: Used to provide member identifier with a
        meaningful prefix.
    """

    def __init__(self, agent_id=None, prefix=''):
        self.coordinator = None
        self.agent_id = agent_id or str(uuid.uuid4())
        self.started = False
        self.prefix = prefix

    @utils.synchronized(name="coordinator_start")
    def start(self):
        if self.started:
            return

        # NOTE(bluex): Tooz expects member_id as a byte string.
        member_id = (self.prefix + self.agent_id).encode('ascii')
        self.coordinator = coordination.get_coordinator(
            cfg.CONF.coordination.backend_url, member_id)
        self.coordinator.start(start_heart=True)
        self.started = True

    def stop(self):
        """Disconnect from coordination backend and stop heartbeat."""
        if self.started:
            self.coordinator.stop()
            self.coordinator = None
            self.started = False

    def get_lock(self, name):
        """Return a Tooz backend lock.

        :param str name: The lock name that is used to identify it
            across all nodes.
        """
        # NOTE(bluex): Tooz expects lock name as a byte string.
        lock_name = (self.prefix + name).encode('ascii')
        if self.coordinator is not None:
            return self.coordinator.get_lock(lock_name)
        else:
            raise exception.NovaException(
                _('Could not create lock. Coordinator uninitialized.'))


COORDINATOR = Coordinator(prefix='compute-hyperv-')


def synchronized(lock_name, blocking=True, coordinator=COORDINATOR):
    """Synchronization decorator.

    :param str lock_name: Lock name.
    :param blocking: If True, blocks until the lock is acquired.
            If False, raises exception when not acquired. Otherwise,
            the value is used as a timeout value and if lock is not acquired
            after this number of seconds exception is raised.
    :param coordinator: Coordinator class to use when creating lock.
        Defaults to the global coordinator.
    :raises tooz.coordination.LockAcquireFailed: if lock is not acquired

    Decorating a method like so::

        @synchronized('mylock')
        def foo(self, *args):
           ...

    ensures that only one process will execute the foo method at a time.

    Different methods can share the same lock::

        @synchronized('mylock')
        def foo(self, *args):
           ...

        @synchronized('mylock')
        def bar(self, *args):
           ...

    This way only one of either foo or bar can be executing at a time.

    Lock name can be formatted using Python format string syntax::

        @synchronized('{f_name}-{vol.id}-{snap[name]}')
        def foo(self, vol, snap):
           ...

    Available field names are: decorated function parameters and
    `f_name` as a decorated function name.
    """

    @decorator.decorator
    def _synchronized(f, *a, **k):
        call_args = inspect.getcallargs(f, *a, **k)
        call_args['f_name'] = f.__name__
        lock = coordinator.get_lock(lock_name.format(**call_args))
        t1 = timeutils.now()
        t2 = None
        try:
            with lock(blocking):
                t2 = timeutils.now()
                LOG.debug('Lock "%(name)s" acquired by "%(function)s" :: '
                          'waited %(wait_secs)0.3fs',
                          {'name': lock.name,
                           'function': f.__name__,
                           'wait_secs': (t2 - t1)})
                return f(*a, **k)
        finally:
            t3 = timeutils.now()
            if t2 is None:
                held_secs = "N/A"
            else:
                held_secs = "%0.3fs" % (t3 - t2)
            LOG.debug('Lock "%(name)s" released by "%(function)s" :: held '
                      '%(held_secs)s',
                      {'name': lock.name,
                       'function': f.__name__,
                       'held_secs': held_secs})

    return _synchronized



# Copyright 2018 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from nova import exception
from nova import objects
from nova.scheduler.client import report
from oslo_log import log as logging

LOG = logging.getLogger(__name__)

CONSUMER_GENERATION_VERSION = "1.28"  # Rocky


class PlacementUtils(object):
    def __init__(self):
        self.reportclient = report.SchedulerReportClient()

    def move_compute_node_allocations(self, context, instance, old_host,
                                      new_host, merge_existing=True):
        LOG.info("Moving instance allocations from compute node %s to %s.",
                 old_host, new_host, instance=instance)

        cn_uuid = objects.ComputeNode.get_by_host_and_nodename(
            context, old_host, old_host).uuid
        new_cn_uuid = objects.ComputeNode.get_by_host_and_nodename(
            context, new_host, new_host).uuid

        self.move_allocations(context, instance.uuid, cn_uuid,
                              new_cn_uuid,
                              merge_existing=merge_existing)

    @report.retries
    def move_allocations(self, context, consumer_uuid, old_rp_uuid,
                         new_rp_uuid, merge_existing=True):
        allocs = self._get_allocs_for_consumer(
            context, consumer_uuid,
            version=CONSUMER_GENERATION_VERSION)
        allocations = allocs['allocations']

        if old_rp_uuid == new_rp_uuid:
            LOG.debug("Requested to move allocations to the "
                      "same provider: %s.", old_rp_uuid)
            return

        if old_rp_uuid not in allocations:
            LOG.warning("Expected to find allocations referencing resource "
                        "provider %s for %s, but found none.",
                        old_rp_uuid, consumer_uuid)
            return

        if merge_existing and new_rp_uuid in allocations:
            LOG.info("Merging existing allocations for consumer %s on "
                     "provider %s: %s.",
                     consumer_uuid, new_rp_uuid, allocations)
            self.merge_resources(
                allocations[new_rp_uuid]['resources'],
                allocations[old_rp_uuid]['resources'])
        else:
            if new_rp_uuid in allocations:
                LOG.info("Replacing existing allocations for consumer %s "
                         "on provider %s: %s",
                         consumer_uuid, new_rp_uuid, allocations)

            allocations[new_rp_uuid] = allocations[old_rp_uuid]

        del allocations[old_rp_uuid]
        self._put_allocs(context, consumer_uuid, allocs,
                         version=CONSUMER_GENERATION_VERSION)

    def _put_allocs(self, context, consumer_uuid, allocations, version=None):
        url = '/allocations/%s' % consumer_uuid
        r = self.reportclient.put(url, allocations,
                                  version=version,
                                  global_request_id=context.global_id)
        if r.status_code != 204:
            errors = r.json().get('errors') or []
            # NOTE(jaypipes): Yes, it sucks doing string comparison like this
            # but we have no error codes, only error messages.
            # TODO(gibi): Use more granular error codes when available
            for err in errors:
                if err.get('code') == 'placement.concurrent_update':
                    reason = (
                        'another process changed the resource providers '
                        'involved in our attempt to put allocations for '
                        'consumer %s' % consumer_uuid)
                    raise report.Retry('put_allocations', reason)
            raise exception.AllocationUpdateFailed(
                        consumer_uuid=consumer_uuid, error=errors)

    def _get_allocs_for_consumer(self, context, consumer, version=None):
        resp = self.reportclient.get('/allocations/%s' % consumer,
                                     version=version,
                                     global_request_id=context.global_id)
        if not resp:
            # TODO(efried): Use code/title/detail to make a better exception
            raise exception.ConsumerAllocationRetrievalFailed(
                consumer_uuid=consumer, error=resp.text)

        return resp.json()

    @staticmethod
    def merge_resources(original_resources, new_resources, sign=1):
        """Merge a list of new resources with existing resources.

        Either add the resources (if sign is 1) or subtract (if sign is -1).
        If the resulting value is 0 do not include the resource in the results.
        """

        all_keys = set(original_resources.keys()) | set(new_resources.keys())
        for key in all_keys:
            value = (original_resources.get(key, 0) +
                     (sign * new_resources.get(key, 0)))
            if value:
                original_resources[key] = value
            else:
                original_resources.pop(key, None)



# Copyright 2018 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt
from nova import context
from nova import exception
from nova import objects
from nova.tests.unit import fake_requests
from oslo_serialization import jsonutils

from compute_hyperv.nova.utils import placement as placement
from compute_hyperv.tests import fake_instance
from compute_hyperv.tests.unit import test_base


@ddt.ddt
class PlacementUtilsTestCase(test_base.HyperVBaseTestCase):
    _autospec_classes = [
        placement.report.SchedulerReportClient
    ]

    _FAKE_PROVIDER = 'fdb5c6d0-e0e9-4411-b952-fb05d6133718'
    _FAKE_RESOURCES = {'VCPU': 1, 'MEMORY_MB': 512, 'DISK_GB': 1}
    _FAKE_ALLOCATIONS = {
        _FAKE_PROVIDER: {'resources': _FAKE_RESOURCES}
    }

    def setUp(self):
        super(PlacementUtilsTestCase, self).setUp()
        self.context = context.get_admin_context()
        self.instance = fake_instance.fake_instance_obj(self.context)

        self.placement = placement.PlacementUtils()
        self.client = self.placement.reportclient

    @mock.patch.object(objects.ComputeNode, 'get_by_host_and_nodename')
    @mock.patch.object(placement.PlacementUtils, 'move_allocations')
    def test_move_compute_node_allocations(self, mock_move_alloc,
                                           mock_get_comp_node):
        mock_get_comp_node.side_effect = [
            mock.Mock(uuid=uuid) for uuid in [mock.sentinel.old_host_uuid,
                                              mock.sentinel.new_host_uuid]]

        self.placement.move_compute_node_allocations(
            self.context, self.instance, mock.sentinel.old_host,
            mock.sentinel.new_host,
            merge_existing=mock.sentinel.merge_existing)

        mock_move_alloc.assert_called_once_with(
            self.context, self.instance.uuid,
            mock.sentinel.old_host_uuid,
            mock.sentinel.new_host_uuid,
            merge_existing=mock.sentinel.merge_existing)
        mock_get_comp_node.assert_has_calls(
            mock.call(self.context, host, host) for host in
            [mock.sentinel.old_host, mock.sentinel.new_host])

    @ddt.data({},  # provider did not change
              {'old_rp': 'fake_rp'})  # provider not included in allocations
    @ddt.unpack
    @mock.patch.object(placement.PlacementUtils, '_get_allocs_for_consumer')
    @mock.patch.object(placement.PlacementUtils, '_put_allocs')
    def test_move_allocations_noop(self, mock_put, mock_get_allocs,
                                   old_rp=_FAKE_PROVIDER,
                                   new_rp=_FAKE_PROVIDER):
        mock_get_allocs.return_value = {'allocations': self._FAKE_ALLOCATIONS}

        self.placement.move_allocations(
            self.context, mock.sentinel.consumer, old_rp, new_rp)

        mock_get_allocs.assert_called_once_with(
            self.context, mock.sentinel.consumer,
            version=placement.CONSUMER_GENERATION_VERSION)
        mock_put.assert_not_called()

    @ddt.data(True, False)
    @mock.patch.object(placement.PlacementUtils, '_get_allocs_for_consumer')
    @mock.patch.object(placement.PlacementUtils, '_put_allocs')
    def test_merge_allocations(self, merge_existing,
                               mock_put, mock_get_allocs):
        old_rp = self._FAKE_PROVIDER
        new_rp = 'new_rp'
        allocs = self._FAKE_ALLOCATIONS.copy()
        allocs[new_rp] = {'resources': self._FAKE_RESOURCES.copy()}

        mock_get_allocs.return_value = {'allocations': allocs}

        if merge_existing:
            exp_resources = {'VCPU': 2, 'MEMORY_MB': 1024, 'DISK_GB': 2}
        else:
            exp_resources = self._FAKE_RESOURCES
        exp_allocs = {new_rp: {'resources': exp_resources}}

        self.placement.move_allocations(
            self.context, mock.sentinel.consumer, old_rp, new_rp,
            merge_existing=merge_existing)

        mock_put.assert_called_once_with(
            self.context, mock.sentinel.consumer,
            {'allocations': exp_allocs},
            version=placement.CONSUMER_GENERATION_VERSION)

    @ddt.data({},  # no errors
              {'status_code': 409,
               'errors': [{'code': 'placement.concurrent_update'}],
               'expected_exc': placement.report.Retry},
              {'status_code': 500,
               'expected_exc': exception.AllocationUpdateFailed})
    @ddt.unpack
    def test_put_allocs(self, status_code=204, expected_exc=None, errors=None):
        response = fake_requests.FakeResponse(
            status_code,
            content=jsonutils.dumps({'errors': errors}))
        self.client.put.return_value = response

        args = (self.context, mock.sentinel.consumer, mock.sentinel.allocs,
                mock.sentinel.version)
        if expected_exc:
            self.assertRaises(expected_exc, self.placement._put_allocs, *args)
        else:
            self.placement._put_allocs(*args)

        self.client.put.assert_called_once_with(
            '/allocations/%s' % mock.sentinel.consumer,
            mock.sentinel.allocs,
            version=mock.sentinel.version,
            global_request_id=self.context.global_id)

    def test_get_allocs(self):
        ret_val = self.placement._get_allocs_for_consumer(
            self.context, mock.sentinel.consumer, mock.sentinel.version)
        exp_val = self.client.get.return_value.json.return_value
        self.assertEqual(exp_val, ret_val)

        self.client.get.assert_called_once_with(
            '/allocations/%s' % mock.sentinel.consumer,
            version=mock.sentinel.version,
            global_request_id=self.context.global_id)

    def test_get_allocs_missing(self):
        self.client.get.return_value = fake_requests.FakeResponse(500)
        self.assertRaises(
            exception.ConsumerAllocationRetrievalFailed,
            self.placement._get_allocs_for_consumer,
            self.context, mock.sentinel.consumer, mock.sentinel.version)

    def test_merge_resources(self):
        resources = {
            'VCPU': 1, 'MEMORY_MB': 1024,
        }
        new_resources = {
            'VCPU': 2, 'MEMORY_MB': 2048, 'CUSTOM_FOO': 1,
        }
        doubled = {
            'VCPU': 3, 'MEMORY_MB': 3072, 'CUSTOM_FOO': 1,
        }
        saved_orig = dict(resources)
        self.placement.merge_resources(resources, new_resources)
        # Check to see that we've doubled our resources
        self.assertEqual(doubled, resources)
        # and then removed those doubled resources
        self.placement.merge_resources(resources, saved_orig, -1)
        self.assertEqual(new_resources, resources)

    def test_merge_resources_zero(self):
        # Test 0 value resources are ignored.
        resources = {
            'VCPU': 1, 'MEMORY_MB': 1024,
        }
        new_resources = {
            'VCPU': 2, 'MEMORY_MB': 2048, 'DISK_GB': 0,
        }
        # The result should not include the zero valued resource.
        doubled = {
            'VCPU': 3, 'MEMORY_MB': 3072,
        }
        self.placement.merge_resources(resources, new_resources)
        self.assertEqual(doubled, resources)

    def test_merge_resources_original_zeroes(self):
        # Confirm that merging that result in a zero in the original
        # excludes the zeroed resource class.
        resources = {
            'VCPU': 3, 'MEMORY_MB': 1023, 'DISK_GB': 1,
        }
        new_resources = {
            'VCPU': 1, 'MEMORY_MB': 512, 'DISK_GB': 1,
        }
        merged = {
            'VCPU': 2, 'MEMORY_MB': 511,
        }
        self.placement.merge_resources(resources, new_resources, -1)
        self.assertEqual(merged, resources)



# Copyright 2018 Cloudbase Solutions Srl
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from nova import exception
from oslo_log import log as logging

from compute_hyperv.nova import constants
from compute_hyperv.nova import volumeops

LOG = logging.getLogger(__name__)


class ClusterVolumeOps(volumeops.VolumeOps):
    def _load_volume_drivers(self):
        self.volume_drivers = {
            constants.STORAGE_PROTOCOL_SMBFS: volumeops.SMBFSVolumeDriver()
        }

    def _get_volume_driver(self, connection_info):
        driver_type = connection_info.get('driver_volume_type')
        if driver_type in [constants.STORAGE_PROTOCOL_ISCSI,
                           constants.STORAGE_PROTOCOL_FC,
                           constants.STORAGE_PROTOCOL_RBD]:
            err_msg = (
                "The Hyper-V Cluster driver does not currently support "
                "passthrough disks (e.g. iSCSI/FC/RBD disks). The reason is "
                "that the volumes need to be available on the destination "
                "host side during an unexpected instance failover. In order "
                "to leverage your storage backend, you may either use the "
                "*standard* Nova Hyper-V driver or use the Cinder SMB volume "
                "driver (which may imply deploying CSVs on top of LUNs "
                "exposed by your storage backend).")
            LOG.error(err_msg)
            raise exception.VolumeDriverNotFound(driver_type=driver_type)

        return super(ClusterVolumeOps, self)._get_volume_driver(
            connection_info)



# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from barbicanclient import client as barbican_client
from keystoneauth1 import session
from nova import exception
from os_win._i18n import _


class PDK(object):

    def create_pdk(self, context, instance, image_meta, pdk_filepath):
        """Generates a pdk file using the barbican container referenced by
        the image metadata or instance metadata. A pdk file is a shielding
        data file which contains a RDP certificate, unattended file,
        volume signature catalogs and guardian metadata.
        """

        with open(pdk_filepath, 'wb') as pdk_file_handle:
            pdk_reference = self._get_pdk_reference(instance, image_meta)
            pdk_container = self._get_pdk_container(context, instance,
                                                    pdk_reference)
            pdk_data = self._get_pdk_data(pdk_container)
            pdk_file_handle.write(pdk_data)

    def _get_pdk_reference(self, instance, image_meta):
        image_pdk_ref = image_meta['properties'].get('img_pdk_reference')
        boot_metadata_pdk_ref = instance.metadata.get('img_pdk_reference')

        if not (image_pdk_ref or boot_metadata_pdk_ref):
            reason = _('A reference to a barbican container containing the '
                       'pdk file must be passed as an image property. This '
                       'is required in order to enable VTPM')
            raise exception.InstanceUnacceptable(instance_id=instance.uuid,
                                                 reason=reason)
        return boot_metadata_pdk_ref or image_pdk_ref

    def _get_pdk_container(self, context, instance, pdk_reference):
        """Retrieves the barbican container containing the pdk file.
        """

        auth = context.get_auth_plugin()
        sess = session.Session(auth=auth)
        brb_client = barbican_client.Client(session=sess)

        try:
            pdk_container = brb_client.containers.get(pdk_reference)
        except Exception as e:
            err_msg = _("Retrieving barbican container with reference "
                        "%(pdk_reference)s failed with error: %(error)s") % {
                        'pdk_reference': pdk_reference,
                        'error': e}
            raise exception.InvalidMetadata(instance_id=instance.uuid,
                                            reason=err_msg)
        return pdk_container

    def _get_pdk_data(self, pdk_container):
        """Return the data from all barbican container's secrets.
        """

        no_of_secrets = len(pdk_container.secrets)
        data = bytes()
        for index in range(no_of_secrets):
            current_secret = pdk_container.secrets[str(index + 1)]
            retrived_secret_data = current_secret.payload
            data += retrived_secret_data
        return data



# Copyright 2015 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import os

from nova import exception
from nova.i18n import _
from nova import utils
from os_win import utilsfactory
from oslo_log import log as logging
from oslo_utils import importutils
import six

from compute_hyperv.nova import pathutils
from compute_hyperv.nova import serialconsolehandler

LOG = logging.getLogger(__name__)

_console_handlers = {}


def instance_synchronized(func):
    @functools.wraps(func)
    def wrapper(self, instance_name, *args, **kwargs):
        @utils.synchronized(instance_name)
        def inner():
            return func(self, instance_name, *args, **kwargs)
        return inner()
    return wrapper


class SerialConsoleOps(object):
    def __init__(self):
        self._vmops_prop = None

        self._vmutils = utilsfactory.get_vmutils()
        self._pathutils = pathutils.PathUtils()

    @property
    def _vmops(self):
        # We have to avoid a circular dependency.
        if not self._vmops_prop:
            self._vmops_prop = importutils.import_class(
                'compute_hyperv.nova.vmops.VMOps')()
        return self._vmops_prop

    @instance_synchronized
    def start_console_handler(self, instance_name):
        if self._vmutils.is_secure_vm(instance_name):
            LOG.warning("Skipping starting serial console handler. "
                        "Shielded/Encrypted VM %(instance_name)s "
                        "doesn't support serial console.",
                        {'instance_name': instance_name})
            return

        # Cleanup existing workers.
        self.stop_console_handler_unsync(instance_name)
        handler = None

        try:
            handler = serialconsolehandler.SerialConsoleHandler(
                instance_name)
            handler.start()
            _console_handlers[instance_name] = handler
        except Exception as exc:
            LOG.error('Instance %(instance_name)s serial console handler '
                      'could not start. Exception %(exc)s',
                      {'instance_name': instance_name, 'exc': exc})
            if handler:
                handler.stop()

    @instance_synchronized
    def stop_console_handler(self, instance_name):
        self.stop_console_handler_unsync(instance_name)

    def stop_console_handler_unsync(self, instance_name):
        handler = _console_handlers.get(instance_name)
        if handler:
            LOG.info("Stopping instance %(instance_name)s "
                     "serial console handler.",
                     {'instance_name': instance_name})
            handler.stop()
            del _console_handlers[instance_name]

    @instance_synchronized
    def get_serial_console(self, instance_name):
        handler = _console_handlers.get(instance_name)
        if not handler:
            raise exception.ConsoleTypeUnavailable(console_type='serial')
        return handler.get_serial_console()

    @instance_synchronized
    def get_console_output(self, instance_name):
        if self._vmutils.is_secure_vm(instance_name):
            err = _("Shielded/Encrypted VMs don't support serial console.")
            raise exception.ConsoleNotAvailable(err)

        console_log_paths = self._pathutils.get_vm_console_log_paths(
            instance_name)

        handler = _console_handlers.get(instance_name)
        if handler:
            handler.flush_console_log()

        try:
            log = b''
            # Start with the oldest console log file.
            for log_path in reversed(console_log_paths):
                if os.path.exists(log_path):
                    with open(log_path, 'rb') as fp:
                        log += fp.read()
            return log
        except IOError as err:
            raise exception.ConsoleLogOutputException(
                instance_id=instance_name, reason=six.text_type(err))

    def start_console_handlers(self):
        active_instances = self._vmutils.get_active_instances()
        for instance_name in active_instances:
            instance_uuid = self._vmops.get_instance_uuid(instance_name)

            if instance_uuid:
                self.start_console_handler(instance_name)
            else:
                LOG.debug("Instance uuid could not be retrieved for "
                          "instance %s. Its serial console output will not "
                          "be handled.", instance_name)



# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from nova import exception
from six.moves import builtins

from compute_hyperv.nova import pdk
from compute_hyperv.tests.unit import test_base


class PDKTestCase(test_base.HyperVBaseTestCase):

    _FAKE_PDK_FILE_PATH = 'C:\\path\\to\\fakepdk.pdk'

    def setUp(self):
        super(PDKTestCase, self).setUp()
        self._pdk = pdk.PDK()

    @mock.patch.object(builtins, 'open')
    @mock.patch.object(pdk.PDK, '_get_pdk_data')
    @mock.patch.object(pdk.PDK, '_get_pdk_container')
    @mock.patch.object(pdk.PDK, '_get_pdk_reference')
    def test_create_pdk(self, mock_get_pdk_reference, mock_get_pdk_container,
                        mock_get_pdk_data, mock_open):
        mock_instance = mock.MagicMock()
        pdk_file_handle = mock_open.return_value.__enter__.return_value

        pdk_reference = mock_get_pdk_reference.return_value
        pdk_container = mock_get_pdk_container.return_value

        self._pdk.create_pdk(mock.sentinel.context,
                             mock_instance,
                             mock.sentinel.image_meta,
                             self._FAKE_PDK_FILE_PATH)
        mock_get_pdk_reference.assert_called_once_with(
            mock_instance, mock.sentinel.image_meta)
        mock_get_pdk_container.assert_called_once_with(mock.sentinel.context,
                                                       mock_instance,
                                                       pdk_reference)
        mock_get_pdk_data.assert_called_once_with(pdk_container)
        pdk_file_handle.write.assert_called_once_with(
            mock_get_pdk_data.return_value)

    def _test_get_pdk_reference(self, pdk_reference=None,
                                image_meta_pdk_ref=None):
        mock_instance = mock.MagicMock(
            metadata={'img_pdk_reference': image_meta_pdk_ref})
        image_meta = {
            'properties': {'img_pdk_reference': pdk_reference}}

        expected_result = image_meta_pdk_ref or pdk_reference
        result = self._pdk._get_pdk_reference(mock_instance,
                                              image_meta)
        self.assertEqual(expected_result, result)

    def test_get_pdk_boot_reference(self):
        self._test_get_pdk_reference(
            image_meta_pdk_ref=mock.sentinel.image_meta_pdk_ref)

    def test_get_pdk_image_reference(self):
        self._test_get_pdk_reference(pdk_reference=mock.sentinel.pdk_reference)

    def test_get_pdk_no_reference(self):
        image_meta = {'properties': {}}
        mock_instance = mock.MagicMock(metadata={})

        self.assertRaises(exception.InstanceUnacceptable,
                          self._pdk._get_pdk_reference,
                          mock_instance, image_meta)

    @mock.patch('barbicanclient.client.Client')
    @mock.patch('keystoneauth1.session.Session')
    def test_get_pdk_container(self, mock_session, mock_barbican_client):
        instance = mock.MagicMock()
        context = mock.MagicMock()
        auth = context.get_auth_plugin.return_value
        sess = mock_session.return_value
        barbican_client = mock_barbican_client.return_value
        barbican_client.containers.get.return_value = (
            mock.sentinel.pdk_container)

        result = self._pdk._get_pdk_container(context, instance,
                                              mock.sentinel.pdk_reference)

        self.assertEqual(mock.sentinel.pdk_container, result)
        mock_session.assert_called_once_with(auth=auth)
        mock_barbican_client.assert_called_once_with(session=sess)

    @mock.patch('barbicanclient.client.Client')
    @mock.patch('keystoneauth1.session.Session')
    def test_get_pdk_container_exception(self, mock_session,
                                         mock_barbican_client):
        instance = mock.MagicMock()
        context = mock.MagicMock()
        auth = context.get_auth_plugin.return_value
        sess = mock_session.return_value

        barbican_client = mock_barbican_client.return_value
        barbican_client.containers.get.side_effect = [
            exception.InvalidMetadata]

        self.assertRaises(exception.InvalidMetadata,
                          self._pdk._get_pdk_container,
                          context,
                          instance,
                          mock.sentinel.pdk_reference)
        mock_session.assert_called_once_with(auth=auth)
        mock_barbican_client.assert_called_once_with(session=sess)

    def test_get_pdk_data(self):
        pdk_container = mock.MagicMock()
        pdk_container.secrets = {'1': mock.MagicMock(payload=b'fake_secret1'),
                                 '2': mock.MagicMock(payload=b'fake_secret2')}

        response = self._pdk._get_pdk_data(pdk_container)
        expected_result = b'fake_secret1fake_secret2'
        self.assertEqual(expected_result, response)



# Copyright (c) 2012, Cloudscaling
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ast
import re

from hacking import core

"""
Guidelines for writing new hacking checks

 - Use only for Nova specific tests. OpenStack general tests
   should be submitted to the common 'hacking' module.
 - Pick numbers in the range N3xx. Find the current test with
   the highest allocated number and then pick the next value.
 - Keep the test method code in the source file ordered based
   on the N3xx value.
 - List the new rule in the top level HACKING.rst file
 - Add test cases for each new rule to nova/tests/unit/test_hacking.py

"""

UNDERSCORE_IMPORT_FILES = []

session_check = re.compile(r"\w*def [a-zA-Z0-9].*[(].*session.*[)]")
cfg_re = re.compile(r".*\scfg\.")
vi_header_re = re.compile(r"^#\s+vim?:.+")
virt_file_re = re.compile(r"\./nova/(?:tests/)?virt/(\w+)/")
virt_import_re = re.compile(
    r"^\s*(?:import|from) nova\.(?:tests\.)?virt\.(\w+)")
virt_config_re = re.compile(
    r"CONF\.import_opt\('.*?', 'nova\.virt\.(\w+)('|.)")
author_tag_re = (re.compile("^\s*#\s*@?(a|A)uthor:"),
                 re.compile("^\.\.\s+moduleauthor::"))
asse_trueinst_re = re.compile(
                     r"(.)*assertTrue\(isinstance\((\w|\.|\'|\"|\[|\])+, "
                     r"(\w|\.|\'|\"|\[|\])+\)\)")
asse_equal_type_re = re.compile(
                       r"(.)*assertEqual\(type\((\w|\.|\'|\"|\[|\])+\), "
                       "(\w|\.|\'|\"|\[|\])+\)")
asse_equal_in_end_with_true_or_false_re = re.compile(r"assertEqual\("
                    r"(\w|[][.'\"])+ in (\w|[][.'\", ])+, (True|False)\)")
asse_equal_in_start_with_true_or_false_re = re.compile(r"assertEqual\("
                    r"(True|False), (\w|[][.'\"])+ in (\w|[][.'\", ])+\)")
asse_equal_end_with_none_re = re.compile(
                           r"assertEqual\(.*?,\s+None\)$")
asse_equal_start_with_none_re = re.compile(
                           r"assertEqual\(None,")
# NOTE(snikitin): Next two regexes weren't united to one for more readability.
#                 asse_true_false_with_in_or_not_in regex checks
#                 assertTrue/False(A in B) cases where B argument has no spaces
#                 asse_true_false_with_in_or_not_in_spaces regex checks cases
#                 where B argument has spaces and starts/ends with [, ', ".
#                 For example: [1, 2, 3], "some string", 'another string'.
#                 We have to separate these regexes to escape a false positives
#                 results. B argument should have spaces only if it starts
#                 with [, ", '. Otherwise checking of string
#                 "assertFalse(A in B and C in D)" will be false positives.
#                 In this case B argument is "B and C in D".
asse_true_false_with_in_or_not_in = re.compile(r"assert(True|False)\("
                    r"(\w|[][.'\"])+( not)? in (\w|[][.'\",])+(, .*)?\)")
asse_true_false_with_in_or_not_in_spaces = re.compile(r"assert(True|False)"
                    r"\((\w|[][.'\"])+( not)? in [\[|'|\"](\w|[][.'\", ])+"
                    r"[\[|'|\"](, .*)?\)")
asse_raises_regexp = re.compile(r"assertRaisesRegexp\(")
conf_attribute_set_re = re.compile(r"CONF\.[a-z0-9_.]+\s*=\s*\w")
translated_log = re.compile(
    r"(.)*LOG\.(audit|debug|error|info|critical|exception|warning)"
    "\(\s*_\(\s*('|\")")
mutable_default_args = re.compile(r"^\s*def .+\((.+=\{\}|.+=\[\])")
string_translation = re.compile(r"[^_]*_\(\s*('|\")")
underscore_import_check = re.compile(r"(.)*import _(.)*")
import_translation_for_log_or_exception = re.compile(
    r"(.)*(from\snova.i18n\simport)\s_")
# We need this for cases where they have created their own _ function.
custom_underscore_check = re.compile(r"(.)*_\s*=\s*(.)*")
api_version_re = re.compile(r"@.*api_version")
dict_constructor_with_list_copy_re = re.compile(r".*\bdict\((\[)?(\(|\[)")
decorator_re = re.compile(r"@.*")

# TODO(dims): When other oslo libraries switch over non-namespace'd
# imports, we need to add them to the regexp below.
oslo_namespace_imports = re.compile(r"from[\s]*oslo[.]"
                                    r"(concurrency|config|db|i18n|messaging|"
                                    r"middleware|serialization|utils|vmware)")
oslo_namespace_imports_2 = re.compile(r"from[\s]*oslo[\s]*import[\s]*"
                                    r"(concurrency|config|db|i18n|messaging|"
                                    r"middleware|serialization|utils|vmware)")
oslo_namespace_imports_3 = re.compile(r"import[\s]*oslo\."
                                    r"(concurrency|config|db|i18n|messaging|"
                                    r"middleware|serialization|utils|vmware)")


class BaseASTChecker(ast.NodeVisitor):
    """Provides a simple framework for writing AST-based checks.

    Subclasses should implement visit_* methods like any other AST visitor
    implementation. When they detect an error for a particular node the
    method should call ``self.add_error(offending_node)``. Details about
    where in the code the error occurred will be pulled from the node
    object.

    Subclasses should also provide a class variable named CHECK_DESC to
    be used for the human readable error message.

    """

    def __init__(self, tree, filename):
        """This object is created automatically by pep8.

        :param tree: an AST tree
        :param filename: name of the file being analyzed
                         (ignored by our checks)
        """
        self._tree = tree
        self._errors = []

    def run(self):
        """Called automatically by pep8."""
        self.visit(self._tree)
        return self._errors

    def add_error(self, node, message=None):
        """Add an error caused by a node to the list of errors for pep8."""
        message = message or self.CHECK_DESC
        error = (node.lineno, node.col_offset, message, self.__class__)
        self._errors.append(error)

    def _check_call_names(self, call_node, names):
        if isinstance(call_node, ast.Call):
            if isinstance(call_node.func, ast.Name):
                if call_node.func.id in names:
                    return True
        return False


@core.flake8ext
def import_no_db_in_virt(logical_line, filename):
    """Check for db calls from nova/virt

    As of grizzly-2 all the database calls have been removed from
    nova/virt, and we want to keep it that way.

    N307
    """
    if "nova/virt" in filename and not filename.endswith("fake.py"):
        if logical_line.startswith("from nova import db"):
            yield (0, "N307: nova.db import not allowed in nova/virt/*")


@core.flake8ext
def no_db_session_in_public_api(logical_line, filename):
    if "db/api.py" in filename:
        if session_check.match(logical_line):
            yield (0, "N309: public db api methods may not accept session")


@core.flake8ext
def use_timeutils_utcnow(logical_line, filename):
    # tools are OK to use the standard datetime module
    if "/tools/" in filename:
        return

    msg = "N310: timeutils.utcnow() must be used instead of datetime.%s()"

    datetime_funcs = ['now', 'utcnow']
    for f in datetime_funcs:
        pos = logical_line.find('datetime.%s' % f)
        if pos != -1:
            yield (pos, msg % f)


def _get_virt_name(regex, data):
    m = regex.match(data)
    if m is None:
        return None
    driver = m.group(1)
    # Ignore things we mis-detect as virt drivers in the regex
    if driver in ["test_virt_drivers", "driver", "firewall",
                  "disk", "api", "imagecache", "cpu", "hardware"]:
        return None
    # TODO(berrange): remove once bugs 1261826 and 126182 are
    # fixed, or baremetal driver is removed, which is first.
    if driver == "baremetal":
        return None
    return driver


@core.flake8ext
def import_no_virt_driver_import_deps(physical_line, filename):
    """Check virt drivers' modules aren't imported by other drivers

    Modules under each virt driver's directory are
    considered private to that virt driver. Other drivers
    in Nova must not access those drivers. Any code that
    is to be shared should be refactored into a common
    module

    N311
    """
    thisdriver = _get_virt_name(virt_file_re, filename)
    thatdriver = _get_virt_name(virt_import_re, physical_line)
    if (thatdriver is not None and
        thisdriver is not None and
        thisdriver != thatdriver):
        return (0, "N311: importing code from other virt drivers forbidden")


@core.flake8ext
def import_no_virt_driver_config_deps(physical_line, filename):
    """Check virt drivers' config vars aren't used by other drivers

    Modules under each virt driver's directory are
    considered private to that virt driver. Other drivers
    in Nova must not use their config vars. Any config vars
    that are to be shared should be moved into a common module

    N312
    """
    thisdriver = _get_virt_name(virt_file_re, filename)
    thatdriver = _get_virt_name(virt_config_re, physical_line)
    if (thatdriver is not None and
        thisdriver is not None and
        thisdriver != thatdriver):
        return (0, "N312: using config vars from other virt drivers forbidden")


@core.flake8ext
def capital_cfg_help(logical_line, tokens):
    msg = "N313: capitalize help string"

    if cfg_re.match(logical_line):
        for t in range(len(tokens)):
            if tokens[t][1] == "help":
                txt = tokens[t + 2][1]
                if len(txt) > 1 and txt[1].islower():
                    yield(0, msg)


@core.flake8ext
def assert_true_instance(logical_line):
    """Check for assertTrue(isinstance(a, b)) sentences

    N316
    """
    if asse_trueinst_re.match(logical_line):
        yield (0, "N316: assertTrue(isinstance(a, b)) sentences not allowed")


@core.flake8ext
def assert_equal_type(logical_line):
    """Check for assertEqual(type(A), B) sentences

    N317
    """
    if asse_equal_type_re.match(logical_line):
        yield (0, "N317: assertEqual(type(A), B) sentences not allowed")


@core.flake8ext
def assert_equal_none(logical_line):
    """Check for assertEqual(A, None) or assertEqual(None, A) sentences

    N318
    """
    res = (asse_equal_start_with_none_re.search(logical_line) or
           asse_equal_end_with_none_re.search(logical_line))
    if res:
        yield (0, "N318: assertEqual(A, None) or assertEqual(None, A) "
               "sentences not allowed")


@core.flake8ext
def no_translate_logs(logical_line):
    """Check for 'LOG.*(_('

    Starting with the Pike series, OpenStack no longer supports log
    translation. We shouldn't translate logs.

    - This check assumes that 'LOG' is a logger.
    - Use filename so we can start enforcing this in specific folders
      instead of needing to do so all at once.

    C312
    """
    if translated_log.match(logical_line):
        yield(0, "C312: Log messages should not be translated!")


@core.flake8ext
def no_import_translation_in_tests(logical_line, filename):
    """Check for 'from nova.i18n import _'
    N337
    """
    if 'nova/tests/' in filename:
        res = import_translation_for_log_or_exception.match(logical_line)
        if res:
            yield(0, "N337 Don't import translation in tests")


@core.flake8ext
def no_setting_conf_directly_in_tests(logical_line, filename):
    """Check for setting CONF.* attributes directly in tests

    The value can leak out of tests affecting how subsequent tests run.
    Using self.flags(option=value) is the preferred method to temporarily
    set config options in tests.

    N320
    """
    if 'nova/tests/' in filename:
        res = conf_attribute_set_re.match(logical_line)
        if res:
            yield (0, "N320: Setting CONF.* attributes directly in tests is "
                      "forbidden. Use self.flags(option=value) instead")


@core.flake8ext
def no_mutable_default_args(logical_line):
    msg = "N322: Method's default argument shouldn't be mutable!"
    if mutable_default_args.match(logical_line):
        yield (0, msg)


@core.flake8ext
def check_explicit_underscore_import(logical_line, filename):
    """Check for explicit import of the _ function

    We need to ensure that any files that are using the _() function
    to translate logs are explicitly importing the _ function.  We
    can't trust unit test to catch whether the import has been
    added so we need to check for it here.
    """

    # Build a list of the files that have _ imported.  No further
    # checking needed once it is found.
    if filename in UNDERSCORE_IMPORT_FILES:
        pass
    elif (underscore_import_check.match(logical_line) or
          custom_underscore_check.match(logical_line)):
        UNDERSCORE_IMPORT_FILES.append(filename)
    elif string_translation.match(logical_line):
        yield(0, "N323: Found use of _() without explicit import of _ !")


@core.flake8ext
def use_jsonutils(logical_line, filename):
    # the code below that path is not meant to be executed from neutron
    # tree where jsonutils module is present, so don't enforce its usage
    # for this subdirectory
    if "plugins/xenserver" in filename:
        return

    # tools are OK to use the standard json module
    if "/tools/" in filename:
        return

    msg = "N324: jsonutils.%(fun)s must be used instead of json.%(fun)s"

    if "json." in logical_line:
        json_funcs = ['dumps(', 'dump(', 'loads(', 'load(']
        for f in json_funcs:
            pos = logical_line.find('json.%s' % f)
            if pos != -1:
                yield (pos, msg % {'fun': f[:-1]})


@core.flake8ext
def check_api_version_decorator(logical_line, previous_logical, blank_before,
                                filename):
    msg = ("N332: the api_version decorator must be the first decorator"
           " on a method.")
    if blank_before == 0 and re.match(api_version_re, logical_line) \
           and re.match(decorator_re, previous_logical):
        yield(0, msg)


class CheckForStrUnicodeExc(BaseASTChecker):
    """Checks for the use of str() or unicode() on an exception.

    This currently only handles the case where str() or unicode()
    is used in the scope of an exception handler.  If the exception
    is passed into a function, returned from an assertRaises, or
    used on an exception created in the same scope, this does not
    catch it.
    """

    name = "check_for_str_unicode_exc"
    version = "1.0"
    CHECK_DESC = ('N325 str() and unicode() cannot be used on an '
                  'exception.  Remove or use six.text_type()')

    def __init__(self, tree, filename):
        super(CheckForStrUnicodeExc, self).__init__(tree, filename)
        self.name = []
        self.already_checked = []

    def visit_TryExcept(self, node):
        for handler in node.handlers:
            if handler.name:
                self.name.append(handler.name.id)
                super(CheckForStrUnicodeExc, self).generic_visit(node)
                self.name = self.name[:-1]
            else:
                super(CheckForStrUnicodeExc, self).generic_visit(node)

    def visit_Call(self, node):
        if self._check_call_names(node, ['str', 'unicode']):
            if node not in self.already_checked:
                self.already_checked.append(node)
                if isinstance(node.args[0], ast.Name):
                    if node.args[0].id in self.name:
                        self.add_error(node.args[0])
        super(CheckForStrUnicodeExc, self).generic_visit(node)


class CheckForTransAdd(BaseASTChecker):
    """Checks for the use of concatenation on a translated string.

    Translations should not be concatenated with other strings, but
    should instead include the string being added to the translated
    string to give the translators the most information.
    """

    name = "check_for_trans_add"
    version = "1.0"
    CHECK_DESC = ('N326 Translated messages cannot be concatenated.  '
                  'String should be included in translated message.')

    TRANS_FUNC = ['_', '_LI', '_LW', '_LE', '_LC']

    def visit_BinOp(self, node):
        if isinstance(node.op, ast.Add):
            if self._check_call_names(node.left, self.TRANS_FUNC):
                self.add_error(node.left)
            elif self._check_call_names(node.right, self.TRANS_FUNC):
                self.add_error(node.right)
        super(CheckForTransAdd, self).generic_visit(node)


@core.flake8ext
def check_oslo_namespace_imports(logical_line, blank_before, filename):
    if re.match(oslo_namespace_imports, logical_line):
        msg = ("N333: '%s' must be used instead of '%s'.") % (
               logical_line.replace('oslo.', 'oslo_'),
               logical_line)
        yield(0, msg)
    match = re.match(oslo_namespace_imports_2, logical_line)
    if match:
        msg = ("N333: 'module %s should not be imported "
               "from oslo namespace.") % match.group(1)
        yield(0, msg)
    match = re.match(oslo_namespace_imports_3, logical_line)
    if match:
        msg = ("N333: 'module %s should not be imported "
               "from oslo namespace.") % match.group(1)
        yield(0, msg)


@core.flake8ext
def assert_true_or_false_with_in(logical_line):
    """Check for assertTrue/False(A in B), assertTrue/False(A not in B),
    assertTrue/False(A in B, message) or assertTrue/False(A not in B, message)
    sentences.

    N334
    """
    res = (asse_true_false_with_in_or_not_in.search(logical_line) or
           asse_true_false_with_in_or_not_in_spaces.search(logical_line))
    if res:
        yield (0, "N334: Use assertIn/NotIn(A, B) rather than "
                  "assertTrue/False(A in/not in B) when checking collection "
                  "contents.")


@core.flake8ext
def assert_raises_regexp(logical_line):
    """Check for usage of deprecated assertRaisesRegexp

    N335
    """
    res = asse_raises_regexp.search(logical_line)
    if res:
        yield (0, "N335: assertRaisesRegex must be used instead "
                  "of assertRaisesRegexp")


@core.flake8ext
def dict_constructor_with_list_copy(logical_line):
    msg = ("N336: Must use a dict comprehension instead of a dict constructor"
           " with a sequence of key-value pairs."
           )
    if dict_constructor_with_list_copy_re.match(logical_line):
        yield (0, msg)


@core.flake8ext
def assert_equal_in(logical_line):
    """Check for assertEqual(A in B, True), assertEqual(True, A in B),
    assertEqual(A in B, False) or assertEqual(False, A in B) sentences

    N338
    """
    res = (asse_equal_in_start_with_true_or_false_re.search(logical_line) or
           asse_equal_in_end_with_true_or_false_re.search(logical_line))
    if res:
        yield (0, "N338: Use assertIn/NotIn(A, B) rather than "
                  "assertEqual(A in B, True/False) when checking collection "
                  "contents.")



# Copyright 2013 Cloudbase Solutions Srl
# Copyright 2013 Pedro Navarro Perez
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from nova import exception
from nova.i18n import _
from nova.network import model
from nova.network import os_vif_util
import os_vif
from os_win import constants as os_win_const
from os_win import utilsfactory
from oslo_log import log as logging

import compute_hyperv.nova.conf

LOG = logging.getLogger(__name__)
CONF = compute_hyperv.nova.conf.CONF


class HyperVBaseVIFPlugin(object):
    @abc.abstractmethod
    def plug(self, instance, vif):
        pass

    @abc.abstractmethod
    def unplug(self, instance, vif):
        pass


class HyperVNeutronVIFPlugin(HyperVBaseVIFPlugin):
    """Neutron VIF plugin."""

    def plug(self, instance, vif):
        # Neutron takes care of plugging the port
        pass

    def unplug(self, instance, vif):
        # Neutron takes care of unplugging the port
        pass


class HyperVNovaNetworkVIFPlugin(HyperVBaseVIFPlugin):
    """Nova network VIF plugin."""

    def __init__(self):
        self._netutils = utilsfactory.get_networkutils()

    def plug(self, instance, vif):
        self._netutils.connect_vnic_to_vswitch(CONF.hyperv.vswitch_name,
                                               vif['id'])

    def unplug(self, instance, vif):
        # TODO(alepilotti) Not implemented
        pass


class HyperVVIFDriver(object):
    def __init__(self):
        self._metricsutils = utilsfactory.get_metricsutils()
        self._netutils = utilsfactory.get_networkutils()
        self._vmutils = utilsfactory.get_vmutils()
        self._vif_plugin = HyperVNeutronVIFPlugin()

    def plug(self, instance, vif):
        vif_type = vif['type']
        if vif_type == model.VIF_TYPE_HYPERV:
            self._vif_plugin.plug(instance, vif)
        elif vif_type == model.VIF_TYPE_OVS:
            reason = _("OVS is no longer supported. Please consider using "
                       "the networking-hyperv agent.")
            raise exception.VirtualInterfacePlugException(reason)
        else:
            reason = _("Failed to plug virtual interface: "
                       "unexpected vif_type=%s") % vif_type
            raise exception.VirtualInterfacePlugException(reason)

    def unplug(self, instance, vif):
        vif_type = vif['type']
        if vif_type == model.VIF_TYPE_HYPERV:
            self._vif_plugin.unplug(instance, vif)
        elif vif_type == model.VIF_TYPE_OVS:
            vif = os_vif_util.nova_to_osvif_vif(vif)
            instance = os_vif_util.nova_to_osvif_instance(instance)
            os_vif.unplug(vif, instance)
        else:
            reason = _("unexpected vif_type=%s") % vif_type
            raise exception.VirtualInterfaceUnplugException(reason=reason)

    def enable_metrics(self, instance_name, vif_id):
        # Hyper-V's metric collection API is extremely inconsistent.
        # As opposed to other metrics, network metrics have to be enabled
        # whenever the vm starts. Attempting to do so while the vm is shut off
        # will fail. Also, this option gets reset when the vm is rebooted.
        #
        # Note that meter ACLs must already be set on the specified port.
        # For "hyperv" ports, this is handled by networking-hyperv, while
        # for OVS ports, we're doing it on the Nova side.
        vm_state = self._vmutils.get_vm_state(instance_name)
        if vm_state in [os_win_const.HYPERV_VM_STATE_ENABLED,
                        os_win_const.HYPERV_VM_STATE_PAUSED]:
            LOG.debug("Enabling instance port metrics. "
                      "Instance name: %(instance_name)s. "
                      "Port name: %(port_name)s.",
                      dict(instance_name=instance_name,
                           port_name=vif_id))
            self._metricsutils.enable_port_metrics_collection(vif_id)
        else:
            LOG.debug("Instance %s is not running. Port metrics will "
                      "be enabled when the instance starts.",
                      instance_name)



# Copyright (c) 2016 Cloudbase Solutions Srl
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
from unittest import mock

import ddt
from nova import block_device
from nova import exception
from nova import objects
from nova.virt import block_device as driver_block_device
from os_win import constants as os_win_const
from os_win import exceptions as os_win_exc
from oslo_serialization import jsonutils

from compute_hyperv.nova import block_device_manager
from compute_hyperv.nova import constants
from compute_hyperv.tests.unit import test_base


@ddt.ddt
class BlockDeviceManagerTestCase(test_base.HyperVBaseTestCase):
    """Unit tests for the Hyper-V BlockDeviceInfoManager class."""

    _autospec_classes = [
        block_device_manager.volumeops.VolumeOps,
        block_device_manager.pathutils.PathUtils,
    ]

    _FAKE_CONN_INFO = {
        'serial': 'fake_volume_id'
    }

    _FAKE_ATTACH_INFO = {
        'controller_type': constants.CTRL_TYPE_SCSI,
        'controller_addr': 0,
        'controller_slot': 1
    }

    def setUp(self):
        super(BlockDeviceManagerTestCase, self).setUp()
        self._bdman = block_device_manager.BlockDeviceInfoManager()

        self._volops = self._bdman._volops
        self._pathutils = self._bdman._pathutils

    @ddt.data(constants.CTRL_TYPE_SCSI, constants.CTRL_TYPE_IDE)
    def test_get_device_bus(self, controller_type):
        fake_ctrl_addr = self._FAKE_ATTACH_INFO['controller_addr']
        fake_ctrl_slot = self._FAKE_ATTACH_INFO['controller_slot']

        bus = self._bdman._get_device_bus(
            controller_type, fake_ctrl_addr, fake_ctrl_slot)

        if controller_type == constants.CTRL_TYPE_SCSI:
            exp_addr = '0:0:%s:%s' % (fake_ctrl_addr, fake_ctrl_slot)
            exp_cls = objects.SCSIDeviceBus
        else:
            exp_addr = '%s:%s' % (fake_ctrl_addr, fake_ctrl_slot)
            exp_cls = objects.IDEDeviceBus

        self.assertIsInstance(bus, exp_cls)
        self.assertEqual(exp_addr, bus.address)

    @ddt.data({},
              {'bdm_is_vol': False},
              {'conn_info_set': False})
    @ddt.unpack
    @mock.patch.object(driver_block_device, 'convert_volume')
    def test_get_vol_bdm_att_info(self, mock_convert_vol,
                                  bdm_is_vol=True,
                                  conn_info_set=True):
        mock_drv_bdm = (dict(connection_info=self._FAKE_CONN_INFO)
                        if conn_info_set else {})
        mock_convert_vol.return_value = (mock_drv_bdm
                                         if bdm_is_vol
                                         else None)

        self._volops.get_disk_attachment_info.return_value = (
            self._FAKE_ATTACH_INFO.copy())

        attach_info = self._bdman._get_vol_bdm_attachment_info(
            mock.sentinel.bdm)

        mock_convert_vol.assert_called_once_with(
            mock.sentinel.bdm)

        if bdm_is_vol and conn_info_set:
            exp_attach_info = self._FAKE_ATTACH_INFO.copy()
            exp_attach_info['serial'] = self._FAKE_CONN_INFO['serial']

            self._volops.get_disk_attachment_info.assert_called_once_with(
                self._FAKE_CONN_INFO)
        else:
            exp_attach_info = None

            self._volops.get_disk_attachment_info.assert_not_called()

        self.assertEqual(exp_attach_info, attach_info)

    @ddt.data({},
              {'eph_name_set': False},
              {'eph_disk_exists': False})
    @ddt.unpack
    @mock.patch.object(block_device_manager.BlockDeviceInfoManager,
                       'get_bdm_connection_info')
    @mock.patch('os.path.exists')
    def test_get_eph_bdm_attachment_info(self, mock_exists,
                                         mock_get_bdm_conn_info,
                                         eph_name_set=True,
                                         eph_disk_exists=True):
        fake_instance_dir = 'fake_instance_dir'
        fake_eph_name = 'eph0.vhdx'
        mock_instance = mock.Mock()

        fake_conn_info = self._FAKE_CONN_INFO.copy()
        if eph_name_set:
            fake_conn_info['eph_filename'] = fake_eph_name

        mock_get_bdm_conn_info.return_value = fake_conn_info
        mock_exists.return_value = eph_disk_exists
        mock_get_attach_info = self._bdman._vmutils.get_disk_attachment_info

        self._pathutils.get_instance_dir.return_value = fake_instance_dir

        attach_info = self._bdman._get_eph_bdm_attachment_info(
            mock_instance, mock.sentinel.bdm)

        if eph_name_set and eph_disk_exists:
            exp_attach_info = mock_get_attach_info.return_value
            exp_eph_path = os.path.join(fake_instance_dir, fake_eph_name)

            mock_exists.assert_called_once_with(exp_eph_path)
            mock_get_attach_info.assert_called_once_with(
                exp_eph_path,
                is_physical=False)
        else:
            exp_attach_info = None

            mock_get_attach_info.assert_not_called()

        self.assertEqual(exp_attach_info, attach_info)

        mock_get_bdm_conn_info.assert_called_once_with(
            mock.sentinel.bdm)

    @mock.patch.object(block_device_manager.BlockDeviceInfoManager,
                       '_get_vol_bdm_attachment_info')
    @mock.patch.object(block_device_manager.BlockDeviceInfoManager,
                       '_get_eph_bdm_attachment_info')
    @mock.patch.object(block_device_manager.BlockDeviceInfoManager,
                       '_get_device_bus')
    @mock.patch.object(block_device, 'new_format_is_ephemeral')
    @mock.patch.object(objects, 'DiskMetadata')
    def test_get_disk_metadata(self, mock_diskmetadata_cls,
                               mock_is_eph,
                               mock_get_device_bus,
                               mock_get_vol_attach_info,
                               mock_get_eph_attach_info,
                               bdm_is_eph=False,
                               bdm_is_vol=False,
                               attach_info_retrieved=True):
        mock_instance = mock.Mock()
        mock_bdm = mock.Mock()
        mock_bdm.is_volume = bdm_is_vol

        if attach_info_retrieved:
            attach_info = self._FAKE_ATTACH_INFO.copy()
            attach_info['serial'] = mock.sentinel.serial
        else:
            attach_info = None

        mock_get_eph_attach_info.return_value = attach_info
        mock_get_vol_attach_info.return_value = attach_info
        mock_is_eph.return_value = bdm_is_eph

        disk_metadata = self._bdman._get_disk_metadata(
            mock_instance, mock_bdm)

        if (bdm_is_vol or bdm_is_eph) and attach_info_retrieved:
            exp_disk_meta = mock_diskmetadata_cls.return_value

            mock_get_device_bus.assert_called_once_with(
                self._FAKE_ATTACH_INFO['controller_type'],
                self._FAKE_ATTACH_INFO['controller_addr'],
                self._FAKE_ATTACH_INFO['controller_slot'])
            mock_diskmetadata_cls.assert_called_once_with(
                bus=mock_get_device_bus.return_value,
                tags=[mock_bdm.tag],
                serial=mock.sentinel.serial)
        else:
            exp_disk_meta = None

            mock_get_device_bus.assert_not_called()

        self.assertEqual(exp_disk_meta, disk_metadata)

        if bdm_is_vol:
            mock_get_vol_attach_info.assert_called_once_with(mock_bdm)
        elif bdm_is_eph:
            mock_get_eph_attach_info.assert_called_once_with(mock_instance,
                                                             mock_bdm)

    @mock.patch.object(block_device_manager.BlockDeviceInfoManager,
                       '_get_disk_metadata')
    @mock.patch.object(objects.BlockDeviceMappingList,
                       'get_by_instance_uuid')
    def test_get_bdm_metadata(self, mock_get_bdm_list,
                              mock_get_disk_meta):
        bdms = [mock.Mock()] * 4
        disk_meta = mock.Mock()
        mock_instance = mock.Mock()

        mock_get_bdm_list.return_value = bdms
        mock_get_disk_meta.side_effect = [
            None,
            exception.DiskNotFound(message='fake_err'),
            os_win_exc.DiskNotFound(message='fake_err'),
            disk_meta]

        bdm_meta = self._bdman.get_bdm_metadata(mock.sentinel.context,
                                                mock_instance)

        self.assertEqual([disk_meta], bdm_meta)

        mock_get_bdm_list.assert_called_once_with(mock.sentinel.context,
                                                  mock_instance.uuid)
        mock_get_disk_meta.assert_has_calls(
            [mock.call(mock_instance, bdm) for bdm in bdms])

    @mock.patch.object(objects.BlockDeviceMapping,
                       'get_by_volume_and_instance')
    def test_set_vol_bdm_conn_info(self, mock_get_bdm):
        mock_instance = mock.Mock()
        mock_bdm = mock_get_bdm.return_value

        self._bdman.set_volume_bdm_connection_info(
            mock.sentinel.context, mock_instance, self._FAKE_CONN_INFO)

        mock_get_bdm.assert_called_once_with(
            mock.sentinel.context,
            self._FAKE_CONN_INFO['serial'],
            mock_instance.uuid)

        self.assertEqual(self._FAKE_CONN_INFO,
                         jsonutils.loads(mock_bdm.connection_info))
        mock_bdm.save.assert_called_once_with()

    def test_get_bdm_connection_info(self):
        bdm = mock.Mock(connection_info=None)
        self.assertEqual({}, self._bdman.get_bdm_connection_info(bdm))

        bdm = mock.Mock()
        bdm.connection_info = jsonutils.dumps(self._FAKE_CONN_INFO)
        self.assertEqual(self._FAKE_CONN_INFO,
                         self._bdman.get_bdm_connection_info(bdm))

    def test_update_bdm_conn_info(self):
        connection_info = self._FAKE_CONN_INFO.copy()

        mock_bdm = mock.Mock()
        mock_bdm.connection_info = jsonutils.dumps(connection_info)

        updates = dict(some_key='some_val',
                       some_other_key='some_other_val')

        self._bdman.update_bdm_connection_info(
            mock_bdm, **updates)

        exp_connection_info = connection_info.copy()
        exp_connection_info.update(**updates)

        self.assertEqual(exp_connection_info,
                         jsonutils.loads(mock_bdm.connection_info))
        mock_bdm.save.assert_called_once_with()

    @mock.patch('nova.virt.configdrive.required_by')
    def test_init_controller_slot_counter_gen1_no_configdrive(
            self, mock_cfg_drive_req):
        mock_cfg_drive_req.return_value = False
        slot_map = self._bdman._initialize_controller_slot_counter(
            mock.sentinel.FAKE_INSTANCE, constants.VM_GEN_1)

        self.assertEqual(slot_map[constants.CTRL_TYPE_IDE][0],
                             os_win_const.IDE_CONTROLLER_SLOTS_NUMBER)
        self.assertEqual(slot_map[constants.CTRL_TYPE_IDE][1],
                         os_win_const.IDE_CONTROLLER_SLOTS_NUMBER)
        self.assertEqual(slot_map[constants.CTRL_TYPE_SCSI][0],
                         os_win_const.SCSI_CONTROLLER_SLOTS_NUMBER)

    @mock.patch('nova.virt.configdrive.required_by')
    def test_init_controller_slot_counter_gen1(self, mock_cfg_drive_req):
        slot_map = self._bdman._initialize_controller_slot_counter(
            mock.sentinel.FAKE_INSTANCE, constants.VM_GEN_1)

        self.assertEqual(slot_map[constants.CTRL_TYPE_IDE][1],
                         os_win_const.IDE_CONTROLLER_SLOTS_NUMBER - 1)

    @mock.patch.object(block_device_manager.configdrive, 'required_by')
    @mock.patch.object(block_device_manager.BlockDeviceInfoManager,
                       '_initialize_controller_slot_counter')
    @mock.patch.object(block_device_manager.BlockDeviceInfoManager,
                       '_check_and_update_root_device')
    @mock.patch.object(block_device_manager.BlockDeviceInfoManager,
                       '_check_and_update_ephemerals')
    @mock.patch.object(block_device_manager.BlockDeviceInfoManager,
                       '_check_and_update_volumes')
    def _check_validate_and_update_bdi(self, mock_check_and_update_vol,
                                       mock_check_and_update_eph,
                                       mock_check_and_update_root,
                                       mock_init_ctrl_cntr,
                                       mock_required_by, available_slots=1):
        mock_required_by.return_value = True
        slot_map = {constants.CTRL_TYPE_SCSI: [available_slots]}
        mock_init_ctrl_cntr.return_value = slot_map

        if available_slots:
            self._bdman.validate_and_update_bdi(mock.sentinel.FAKE_INSTANCE,
                                                mock.sentinel.IMAGE_META,
                                                constants.VM_GEN_2,
                                                mock.sentinel.BLOCK_DEV_INFO)
        else:
            self.assertRaises(exception.InvalidBDMFormat,
                              self._bdman.validate_and_update_bdi,
                              mock.sentinel.FAKE_INSTANCE,
                              mock.sentinel.IMAGE_META,
                              constants.VM_GEN_2,
                              mock.sentinel.BLOCK_DEV_INFO)

        mock_init_ctrl_cntr.assert_called_once_with(
            mock.sentinel.FAKE_INSTANCE, constants.VM_GEN_2)
        mock_check_and_update_root.assert_called_once_with(
            constants.VM_GEN_2, mock.sentinel.IMAGE_META,
            mock.sentinel.BLOCK_DEV_INFO, slot_map)
        mock_check_and_update_eph.assert_called_once_with(
            constants.VM_GEN_2, mock.sentinel.BLOCK_DEV_INFO, slot_map)
        mock_check_and_update_vol.assert_called_once_with(
            constants.VM_GEN_2, mock.sentinel.BLOCK_DEV_INFO, slot_map)
        mock_required_by.assert_called_once_with(mock.sentinel.FAKE_INSTANCE)

    def test_validate_and_update_bdi(self):
        self._check_validate_and_update_bdi()

    def test_validate_and_update_bdi_insufficient_slots(self):
        self._check_validate_and_update_bdi(available_slots=0)

    @mock.patch.object(block_device_manager.BlockDeviceInfoManager,
                       '_get_available_controller_slot')
    @mock.patch.object(block_device_manager.BlockDeviceInfoManager,
                       'is_boot_from_volume')
    def _test_check_and_update_root_device(self, mock_is_boot_from_vol,
                                           mock_get_avail_ctrl_slot,
                                           disk_format,
                                           vm_gen=constants.VM_GEN_1,
                                           boot_from_volume=False):
        image_meta = {'disk_format': disk_format}
        bdi = {'root_device': '/dev/sda',
               'block_device_mapping': [
                    {'mount_device': '/dev/sda',
                     'connection_info': mock.sentinel.FAKE_CONN_INFO}]}

        mock_is_boot_from_vol.return_value = boot_from_volume
        mock_get_avail_ctrl_slot.return_value = (0, 0)

        self._bdman._check_and_update_root_device(vm_gen, image_meta, bdi,
                                                  mock.sentinel.SLOT_MAP)

        root_disk = bdi['root_disk']
        if boot_from_volume:
            self.assertEqual(root_disk['type'], constants.VOLUME)
            self.assertIsNone(root_disk['path'])
            self.assertEqual(root_disk['connection_info'],
                             mock.sentinel.FAKE_CONN_INFO)
        else:
            image_type = self._bdman._TYPE_FOR_DISK_FORMAT.get(
                image_meta['disk_format'])
            self.assertEqual(root_disk['type'], image_type)
            self.assertIsNone(root_disk['path'])
            self.assertIsNone(root_disk['connection_info'])

        disk_bus = (constants.CTRL_TYPE_IDE if
            vm_gen == constants.VM_GEN_1 else constants.CTRL_TYPE_SCSI)
        self.assertEqual(root_disk['disk_bus'], disk_bus)
        self.assertEqual(root_disk['drive_addr'], 0)
        self.assertEqual(root_disk['ctrl_disk_addr'], 0)
        self.assertEqual(root_disk['boot_index'], 0)
        self.assertEqual(root_disk['mount_device'], bdi['root_device'])
        mock_get_avail_ctrl_slot.assert_called_once_with(
            root_disk['disk_bus'], mock.sentinel.SLOT_MAP)

    @mock.patch.object(block_device_manager.BlockDeviceInfoManager,
                       'is_boot_from_volume', return_value=False)
    def test_check_and_update_root_device_exception(self, mock_is_boot_vol):
        bdi = {}
        image_meta = mock.MagicMock(disk_format=mock.sentinel.fake_format)

        self.assertRaises(exception.InvalidImageFormat,
                          self._bdman._check_and_update_root_device,
                          constants.VM_GEN_1, image_meta, bdi,
                          mock.sentinel.SLOT_MAP)

    def test_check_and_update_root_device_gen1(self):
        self._test_check_and_update_root_device(disk_format='vhd')

    def test_check_and_update_root_device_gen1_vhdx(self):
        self._test_check_and_update_root_device(disk_format='vhdx')

    def test_check_and_update_root_device_gen1_iso(self):
        self._test_check_and_update_root_device(disk_format='iso')

    def test_check_and_update_root_device_gen2(self):
        self._test_check_and_update_root_device(disk_format='vhd',
                                                vm_gen=constants.VM_GEN_2)

    def test_check_and_update_root_device_boot_from_vol_gen1(self):
        self._test_check_and_update_root_device(disk_format='vhd',
                                                boot_from_volume=True)

    def test_check_and_update_root_device_boot_from_vol_gen2(self):
        self._test_check_and_update_root_device(disk_format='vhd',
                                                vm_gen=constants.VM_GEN_2,
                                                boot_from_volume=True)

    @mock.patch('nova.virt.configdrive.required_by', return_value=True)
    def _test_get_available_controller_slot(self, mock_config_drive_req,
                                            bus=constants.CTRL_TYPE_IDE,
                                            fail=False):

        slot_map = self._bdman._initialize_controller_slot_counter(
            mock.sentinel.FAKE_VM, constants.VM_GEN_1)

        if fail:
            slot_map[constants.CTRL_TYPE_IDE][0] = 0
            slot_map[constants.CTRL_TYPE_IDE][1] = 0
            self.assertRaises(exception.InvalidBDMFormat,
                              self._bdman._get_available_controller_slot,
                              constants.CTRL_TYPE_IDE,
                              slot_map)
        else:
            (disk_addr,
             ctrl_disk_addr) = self._bdman._get_available_controller_slot(
                bus, slot_map)

            self.assertEqual(0, disk_addr)
            self.assertEqual(0, ctrl_disk_addr)

    def test_get_available_controller_slot(self):
        self._test_get_available_controller_slot()

    def test_get_available_controller_slot_scsi_ctrl(self):
        self._test_get_available_controller_slot(bus=constants.CTRL_TYPE_SCSI)

    def test_get_available_controller_slot_exception(self):
        self._test_get_available_controller_slot(fail=True)

    def test_is_boot_from_volume_true(self):
        vol = {'mount_device': self._bdman._DEFAULT_ROOT_DEVICE}
        block_device_info = {'block_device_mapping': [vol]}
        ret = self._bdman.is_boot_from_volume(block_device_info)

        self.assertTrue(ret)

    def test_is_boot_from_volume_false(self):
        block_device_info = {'block_device_mapping': []}
        ret = self._bdman.is_boot_from_volume(block_device_info)

        self.assertFalse(ret)

    def test_get_root_device_bdm(self):
        mount_device = '/dev/sda'
        bdm1 = {'mount_device': None}
        bdm2 = {'mount_device': mount_device}
        bdi = {'block_device_mapping': [bdm1, bdm2]}

        ret = self._bdman._get_root_device_bdm(bdi, mount_device)

        self.assertEqual(bdm2, ret)

    @mock.patch.object(block_device_manager.BlockDeviceInfoManager,
                       '_check_and_update_bdm')
    def test_check_and_update_ephemerals(self, mock_check_and_update_bdm):
        fake_ephemerals = [mock.sentinel.eph1, mock.sentinel.eph2,
                           mock.sentinel.eph3]
        fake_bdi = {'ephemerals': fake_ephemerals}
        expected_calls = []
        for eph in fake_ephemerals:
            expected_calls.append(mock.call(mock.sentinel.fake_slot_map,
                                            mock.sentinel.fake_vm_gen,
                                            eph))
        self._bdman._check_and_update_ephemerals(mock.sentinel.fake_vm_gen,
                                                 fake_bdi,
                                                 mock.sentinel.fake_slot_map)
        mock_check_and_update_bdm.assert_has_calls(expected_calls)

    @mock.patch.object(block_device_manager.BlockDeviceInfoManager,
                       '_check_and_update_bdm')
    @mock.patch.object(block_device_manager.BlockDeviceInfoManager,
                       '_get_root_device_bdm')
    def test_check_and_update_volumes(self, mock_get_root_dev_bdm,
                                      mock_check_and_update_bdm):
        fake_vol1 = {'mount_device': '/dev/sda'}
        fake_vol2 = {'mount_device': '/dev/sdb'}
        fake_volumes = [fake_vol1, fake_vol2]
        fake_bdi = {'block_device_mapping': fake_volumes,
                    'root_disk': {'mount_device': '/dev/sda'}}
        mock_get_root_dev_bdm.return_value = fake_vol1

        self._bdman._check_and_update_volumes(mock.sentinel.fake_vm_gen,
                                              fake_bdi,
                                              mock.sentinel.fake_slot_map)

        mock_get_root_dev_bdm.assert_called_once_with(fake_bdi, '/dev/sda')
        mock_check_and_update_bdm.assert_called_once_with(
            mock.sentinel.fake_slot_map, mock.sentinel.fake_vm_gen, fake_vol2)
        self.assertNotIn(fake_vol1, fake_bdi)

    @mock.patch.object(block_device_manager.BlockDeviceInfoManager,
                       '_get_available_controller_slot')
    def test_check_and_update_bdm_with_defaults(self, mock_get_ctrl_slot):
        mock_get_ctrl_slot.return_value = ((mock.sentinel.DRIVE_ADDR,
                                            mock.sentinel.CTRL_DISK_ADDR))
        bdm = {'device_type': None,
               'disk_bus': None,
               'boot_index': None}

        self._bdman._check_and_update_bdm(mock.sentinel.FAKE_SLOT_MAP,
                                          constants.VM_GEN_1, bdm)

        mock_get_ctrl_slot.assert_called_once_with(
          bdm['disk_bus'], mock.sentinel.FAKE_SLOT_MAP)
        self.assertEqual(mock.sentinel.DRIVE_ADDR, bdm['drive_addr'])
        self.assertEqual(mock.sentinel.CTRL_DISK_ADDR, bdm['ctrl_disk_addr'])
        self.assertEqual('disk', bdm['device_type'])
        self.assertEqual(self._bdman._DEFAULT_BUS, bdm['disk_bus'])
        self.assertIsNone(bdm['boot_index'])

    def test_check_and_update_bdm_exception_device_type(self):
        bdm = {'device_type': 'cdrom',
               'disk_bus': 'IDE'}

        self.assertRaises(exception.InvalidDiskInfo,
                          self._bdman._check_and_update_bdm,
                          mock.sentinel.FAKE_SLOT_MAP, constants.VM_GEN_1, bdm)

    def test_check_and_update_bdm_exception_disk_bus(self):
        bdm = {'device_type': 'disk',
               'disk_bus': 'fake_bus'}

        self.assertRaises(exception.InvalidDiskInfo,
                          self._bdman._check_and_update_bdm,
                          mock.sentinel.FAKE_SLOT_MAP, constants.VM_GEN_1, bdm)

    def test_sort_by_boot_order(self):
        original = [{'boot_index': 2}, {'boot_index': None}, {'boot_index': 1}]
        expected = [original[2], original[0], original[1]]

        self._bdman._sort_by_boot_order(original)
        self.assertEqual(expected, original)

    @mock.patch.object(block_device_manager.BlockDeviceInfoManager,
                       '_get_boot_order_gen1')
    def test_get_boot_order_gen1_vm(self, mock_get_boot_order):
        self._bdman.get_boot_order(constants.VM_GEN_1,
                                   mock.sentinel.BLOCK_DEV_INFO)
        mock_get_boot_order.assert_called_once_with(
            mock.sentinel.BLOCK_DEV_INFO)

    @mock.patch.object(block_device_manager.BlockDeviceInfoManager,
                       '_get_boot_order_gen2')
    def test_get_boot_order_gen2_vm(self, mock_get_boot_order):
        self._bdman.get_boot_order(constants.VM_GEN_2,
                                   mock.sentinel.BLOCK_DEV_INFO)
        mock_get_boot_order.assert_called_once_with(
            mock.sentinel.BLOCK_DEV_INFO)

    def test_get_boot_order_gen1_iso(self):
        fake_bdi = {'root_disk': {'type': 'iso'}}
        expected = [os_win_const.BOOT_DEVICE_CDROM,
                    os_win_const.BOOT_DEVICE_HARDDISK,
                    os_win_const.BOOT_DEVICE_NETWORK,
                    os_win_const.BOOT_DEVICE_FLOPPY]

        res = self._bdman._get_boot_order_gen1(fake_bdi)
        self.assertEqual(expected, res)

    def test_get_boot_order_gen1_vhd(self):
        fake_bdi = {'root_disk': {'type': 'vhd'}}
        expected = [os_win_const.BOOT_DEVICE_HARDDISK,
                    os_win_const.BOOT_DEVICE_CDROM,
                    os_win_const.BOOT_DEVICE_NETWORK,
                    os_win_const.BOOT_DEVICE_FLOPPY]

        res = self._bdman._get_boot_order_gen1(fake_bdi)
        self.assertEqual(expected, res)

    def test_get_boot_order_gen2(self):
        fake_root_disk = {'boot_index': 0,
                          'path': mock.sentinel.FAKE_ROOT_PATH}
        fake_eph1 = {'boot_index': 2,
                     'path': mock.sentinel.FAKE_EPH_PATH1}
        fake_eph2 = {'boot_index': 3,
                     'path': mock.sentinel.FAKE_EPH_PATH2}
        fake_bdm = {'boot_index': 1,
                    'connection_info': mock.sentinel.FAKE_CONN_INFO}
        fake_bdi = {'root_disk': fake_root_disk,
                    'ephemerals': [fake_eph1,
                                   fake_eph2],
                    'block_device_mapping': [fake_bdm]}

        self._bdman._volops.get_disk_resource_path = (
            mock.MagicMock(return_value=fake_bdm['connection_info']))

        expected_res = [mock.sentinel.FAKE_ROOT_PATH,
                        mock.sentinel.FAKE_CONN_INFO,
                        mock.sentinel.FAKE_EPH_PATH1,
                        mock.sentinel.FAKE_EPH_PATH2]

        res = self._bdman._get_boot_order_gen2(fake_bdi)

        self.assertEqual(expected_res, res)



#  Copyright 2014 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
from unittest import mock

import ddt
from nova import block_device
from nova import exception
from nova.virt import driver
from os_win import exceptions as os_win_exc
from oslo_utils import units

from compute_hyperv.nova import constants
from compute_hyperv.nova import migrationops
from compute_hyperv.tests import fake_instance
from compute_hyperv.tests.unit import test_base


@ddt.ddt
class MigrationOpsTestCase(test_base.HyperVBaseTestCase):
    """Unit tests for the Hyper-V MigrationOps class."""

    _autospec_classes = [
        migrationops.pathutils.PathUtils,
        migrationops.volumeops.VolumeOps,
        migrationops.vmops.VMOps,
        migrationops.imagecache.ImageCache,
        migrationops.block_device_manager.BlockDeviceInfoManager,
    ]

    _FAKE_DISK = 'fake_disk'
    _FAKE_TIMEOUT = 10
    _FAKE_RETRY_INTERVAL = 5

    def setUp(self):
        super(MigrationOpsTestCase, self).setUp()
        self.context = 'fake-context'

        self._migrationops = migrationops.MigrationOps()
        self._vmops = self._migrationops._vmops
        self._vmutils = self._migrationops._vmutils
        self._pathutils = self._migrationops._pathutils
        self._vhdutils = self._migrationops._vhdutils
        self._volumeops = self._migrationops._volumeops
        self._imagecache = self._migrationops._imagecache
        self._block_dev_man = self._migrationops._block_dev_man

    def test_move_vm_files(self):
        mock_instance = fake_instance.fake_instance_obj(self.context)

        vm_files_path = self._migrationops._move_vm_files(mock_instance)

        mock_get_inst_dir = self._migrationops._pathutils.get_instance_dir
        mock_get_inst_dir.assert_called_once_with(mock_instance.name)
        mock_get_revert_dir = (
            self._migrationops._pathutils.get_instance_migr_revert_dir)
        mock_get_revert_dir.assert_called_once_with(
            mock_get_inst_dir.return_value, remove_dir=True, create_dir=True)
        mock_get_export_dir = self._migrationops._pathutils.get_export_dir
        mock_get_export_dir.assert_called_once_with(
            instance_dir=mock_get_revert_dir.return_value, create_dir=True)

        mock_move = self._migrationops._pathutils.move_folder_files
        mock_move.assert_called_once_with(mock_get_inst_dir.return_value,
                                          mock_get_revert_dir.return_value)
        copy_config_files = self._migrationops._pathutils.copy_vm_config_files
        copy_config_files.assert_called_once_with(
            mock_instance.name, mock_get_export_dir.return_value)
        self.assertEqual(mock_get_revert_dir.return_value, vm_files_path)

    @ddt.data({},
              {'ephemerals_size': 2},
              {'ephemerals_size': 3, 'flavor_eph_size': 0},
              {'ephemerals_size': 3, 'expect_invalid_flavor': True},
              {'current_root_gb': 3, 'expect_invalid_flavor': True},
              {'current_root_gb': 3, 'boot_from_vol': True})
    @ddt.unpack
    @mock.patch.object(driver, 'block_device_info_get_ephemerals')
    @mock.patch.object(block_device, 'get_bdm_ephemeral_disk_size')
    def test_check_target_flavor(self, mock_get_eph_size, mock_get_eph,
                                 ephemerals_size=0,
                                 flavor_eph_size=2,
                                 flavor_root_gb=2,
                                 current_root_gb=1,
                                 boot_from_vol=False,
                                 expect_invalid_flavor=False):
        mock_instance = fake_instance.fake_instance_obj(self.context)
        mock_instance.flavor.root_gb = current_root_gb
        mock_flavor = mock.MagicMock(root_gb=flavor_root_gb,
                                     ephemeral_gb=flavor_eph_size)

        mock_get_eph_size.return_value = ephemerals_size
        self._block_dev_man.is_boot_from_volume.return_value = boot_from_vol

        if expect_invalid_flavor:
            self.assertRaises(exception.InstanceFaultRollback,
                              self._migrationops._check_target_flavor,
                              mock_instance, mock_flavor,
                              mock.sentinel.block_device_info)
        else:
            self._migrationops._check_target_flavor(
                mock_instance, mock_flavor, mock.sentinel.block_device_info)

        mock_get_eph.assert_called_once_with(mock.sentinel.block_device_info)
        mock_get_eph_size.assert_called_once_with(mock_get_eph.return_value)
        self._block_dev_man.is_boot_from_volume.assert_called_once_with(
            mock.sentinel.block_device_info)

    def test_check_and_attach_config_drive(self):
        mock_instance = fake_instance.fake_instance_obj(
            self.context, expected_attrs=['system_metadata'])
        mock_instance.config_drive = 'True'

        self._migrationops._check_and_attach_config_drive(
            mock_instance, mock.sentinel.vm_gen)

        self._migrationops._vmops.attach_config_drive.assert_called_once_with(
            mock_instance,
            self._migrationops._pathutils.lookup_configdrive_path.return_value,
            mock.sentinel.vm_gen)

    def test_check_and_attach_config_drive_unknown_path(self):
        instance = fake_instance.fake_instance_obj(
            self.context, expected_attrs=['system_metadata'])
        instance.config_drive = 'True'
        self._migrationops._pathutils.lookup_configdrive_path.return_value = (
            None)
        self.assertRaises(exception.ConfigDriveNotFound,
                          self._migrationops._check_and_attach_config_drive,
                          instance,
                          mock.sentinel.FAKE_VM_GEN)

    @mock.patch.object(migrationops.MigrationOps, '_move_vm_files')
    @mock.patch.object(migrationops.MigrationOps, '_check_target_flavor')
    def test_migrate_disk_and_power_off(self, mock_check_flavor,
                                        mock_move_vm_files):
        instance = mock.MagicMock()
        instance.system_metadata = {}
        flavor = mock.MagicMock()
        network_info = mock.MagicMock()

        disk_info = self._migrationops.migrate_disk_and_power_off(
            self.context, instance, mock.sentinel.FAKE_DEST, flavor,
            network_info, mock.sentinel.bdi,
            self._FAKE_TIMEOUT, self._FAKE_RETRY_INTERVAL)

        self.assertEqual(mock_move_vm_files.return_value, disk_info)
        mock_check_flavor.assert_called_once_with(
            instance, flavor, mock.sentinel.bdi)
        self._migrationops._vmops.power_off.assert_called_once_with(
            instance, self._FAKE_TIMEOUT, self._FAKE_RETRY_INTERVAL)
        mock_move_vm_files.assert_called_once_with(instance)
        self.assertEqual(mock_move_vm_files.return_value,
                         instance.system_metadata['backup_location'])
        instance.save.assert_called_once_with()
        self._migrationops._vmops.destroy.assert_called_once_with(
            instance, network_info, mock.sentinel.bdi, destroy_disks=True,
            cleanup_migration_files=False)

    def test_confirm_migration(self):
        mock_instance = fake_instance.fake_instance_obj(
            self.context, expected_attrs=['system_metadata'])
        fake_path_revert = 'fake_path_revert'
        mock_instance.system_metadata['backup_location'] = fake_path_revert

        self._migrationops.confirm_migration(
            context=self.context,
            migration=mock.sentinel.migration, instance=mock_instance,
            network_info=mock.sentinel.network_info)

        get_export_dir = self._migrationops._pathutils.get_export_dir
        get_export_dir.assert_called_once_with(instance_dir=fake_path_revert)
        self._migrationops._pathutils.check_dir.assert_has_calls([
            mock.call(get_export_dir.return_value, remove_dir=True),
            mock.call(fake_path_revert, remove_dir=True)])

    def test_revert_migration_files(self):
        mock_instance = fake_instance.fake_instance_obj(
            self.context, expected_attrs=['system_metadata'])
        fake_path_revert = 'fake_path_revert'
        mock_instance.system_metadata['backup_location'] = fake_path_revert

        instance_path = self._migrationops._revert_migration_files(
            mock_instance)

        expected_instance_path = fake_path_revert.rstrip('_revert')
        self.assertEqual(expected_instance_path, instance_path)
        self._migrationops._pathutils.rename.assert_called_once_with(
            fake_path_revert, expected_instance_path)

    @mock.patch.object(migrationops.MigrationOps, '_import_and_setup_vm')
    @mock.patch.object(migrationops.MigrationOps, '_revert_migration_files')
    def test_finish_revert_migration(self, mock_revert_migration_files,
                                     mock_import_and_setup_vm):
        mock_instance = fake_instance.fake_instance_obj(self.context)

        self._migrationops.finish_revert_migration(
            context=self.context, instance=mock_instance,
            network_info=mock.sentinel.network_info,
            block_device_info=mock.sentinel.block_device_info,
            power_on=True)

        mock_revert_migration_files.assert_called_once_with(
            mock_instance)
        image_meta = self._imagecache.get_image_details.return_value
        mock_import_and_setup_vm.assert_called_once_with(
            self.context, mock_instance,
            mock_revert_migration_files.return_value,
            image_meta, mock.sentinel.block_device_info)
        self._migrationops._vmops.power_on.assert_called_once_with(
            mock_instance, network_info=mock.sentinel.network_info)

    def test_merge_base_vhd(self):
        fake_diff_vhd_path = 'fake/diff/path'
        fake_base_vhd_path = 'fake/base/path'
        base_vhd_copy_path = os.path.join(
            os.path.dirname(fake_diff_vhd_path),
            os.path.basename(fake_base_vhd_path))

        self._migrationops._merge_base_vhd(diff_vhd_path=fake_diff_vhd_path,
                                           base_vhd_path=fake_base_vhd_path)

        self._migrationops._pathutils.copyfile.assert_called_once_with(
            fake_base_vhd_path, base_vhd_copy_path)
        recon_parent_vhd = self._migrationops._vhdutils.reconnect_parent_vhd
        recon_parent_vhd.assert_called_once_with(fake_diff_vhd_path,
                                                 base_vhd_copy_path)
        self._migrationops._vhdutils.merge_vhd.assert_called_once_with(
            fake_diff_vhd_path)
        self._migrationops._pathutils.rename.assert_called_once_with(
            base_vhd_copy_path, fake_diff_vhd_path)

    def test_merge_base_vhd_exception(self):
        fake_diff_vhd_path = 'fake/diff/path'
        fake_base_vhd_path = 'fake/base/path'
        base_vhd_copy_path = os.path.join(
            os.path.dirname(fake_diff_vhd_path),
            os.path.basename(fake_base_vhd_path))

        self._migrationops._vhdutils.reconnect_parent_vhd.side_effect = (
            os_win_exc.HyperVException)
        self._migrationops._pathutils.exists.return_value = True

        self.assertRaises(os_win_exc.HyperVException,
                          self._migrationops._merge_base_vhd,
                          fake_diff_vhd_path, fake_base_vhd_path)
        self._migrationops._pathutils.exists.assert_called_once_with(
            base_vhd_copy_path)
        self._migrationops._pathutils.remove.assert_called_once_with(
            base_vhd_copy_path)

    @mock.patch.object(migrationops.MigrationOps, '_resize_vhd')
    def test_check_resize_vhd(self, mock_resize_vhd):
        self._migrationops._check_resize_vhd(
            vhd_path=mock.sentinel.vhd_path, vhd_info={'VirtualSize': 1},
            new_size=2)
        mock_resize_vhd.assert_called_once_with(mock.sentinel.vhd_path, 2)

    def test_check_resize_vhd_exception(self):
        self.assertRaises(exception.CannotResizeDisk,
                          self._migrationops._check_resize_vhd,
                          mock.sentinel.vhd_path,
                          {'VirtualSize': 1}, 0)

    @mock.patch.object(migrationops.MigrationOps, '_merge_base_vhd')
    def test_resize_vhd(self, mock_merge_base_vhd):
        fake_vhd_path = 'fake/path.vhd'
        new_vhd_size = 2
        self._migrationops._resize_vhd(vhd_path=fake_vhd_path,
                                       new_size=new_vhd_size)

        get_vhd_parent_path = self._migrationops._vhdutils.get_vhd_parent_path
        get_vhd_parent_path.assert_called_once_with(fake_vhd_path)
        mock_merge_base_vhd.assert_called_once_with(
            fake_vhd_path,
            self._migrationops._vhdutils.get_vhd_parent_path.return_value)
        self._migrationops._vhdutils.resize_vhd.assert_called_once_with(
            fake_vhd_path, new_vhd_size)

    def test_check_base_disk(self):
        mock_instance = fake_instance.fake_instance_obj(self.context)
        fake_src_vhd_path = 'fake/src/path'
        fake_base_vhd = 'fake/vhd'
        get_cached_image = self._migrationops._imagecache.get_cached_image
        get_cached_image.return_value = fake_base_vhd

        self._migrationops._check_base_disk(
            context=self.context, instance=mock_instance,
            diff_vhd_path=mock.sentinel.diff_vhd_path,
            src_base_disk_path=fake_src_vhd_path)

        get_cached_image.assert_called_once_with(self.context, mock_instance)
        recon_parent_vhd = self._migrationops._vhdutils.reconnect_parent_vhd
        recon_parent_vhd.assert_called_once_with(
            mock.sentinel.diff_vhd_path, fake_base_vhd)

    @ddt.data((False, '\\\\fake-srv\\C$\\inst_dir_0000000e_revert', True),
              (False, '\\\\fake-srv\\share_path\\inst_dir_0000000e_revert'),
              (True, 'C:\\fake_inst_dir_0000000e_revert'))
    @ddt.unpack
    def test_migrate_disks_from_source(self, move_disks_on_migration,
                                       source_inst_dir, is_remote_path=False):
        self.flags(move_disks_on_cold_migration=move_disks_on_migration,
                   group='hyperv')
        mock_migration = mock.MagicMock()
        mock_instance = fake_instance.fake_instance_obj(self.context)
        mock_get_remote_path = self._migrationops._pathutils.get_remote_path
        mock_get_remote_path.return_value = source_inst_dir

        mock_get_export_dir = self._migrationops._pathutils.get_export_dir
        mock_get_export_dir.side_effect = [mock.sentinel.source_export_dir,
                                           mock.sentinel.dest_export_dir]

        instance_dir = self._migrationops._migrate_disks_from_source(
            mock_migration, mock_instance, mock.sentinel.source_dir)

        mock_get_remote_path.assert_called_once_with(
            mock_migration.source_compute, mock.sentinel.source_dir)

        if move_disks_on_migration or is_remote_path:
            mock_get_inst_dir = self._migrationops._pathutils.get_instance_dir
            mock_get_inst_dir.assert_called_once_with(
                mock_instance.name, create_dir=True, remove_dir=True)
            expected_inst_dir = mock_get_inst_dir.return_value
        else:
            expected_inst_dir = source_inst_dir[0: - len('_revert')]
            self._migrationops._pathutils.check_dir.assert_called_once_with(
                expected_inst_dir, create_dir=True)

        mock_get_export_dir.assert_has_calls([
            mock.call(instance_dir=mock_get_remote_path.return_value),
            mock.call(instance_dir=expected_inst_dir)])

        mock_copy = self._migrationops._pathutils.copy_folder_files
        mock_copy.assert_called_once_with(mock_get_remote_path.return_value,
                                          expected_inst_dir)
        self._migrationops._pathutils.copy_dir.assert_called_once_with(
            mock.sentinel.source_export_dir, mock.sentinel.dest_export_dir)
        self.assertEqual(expected_inst_dir, instance_dir)

    @mock.patch.object(migrationops.MigrationOps, '_import_and_setup_vm')
    @mock.patch.object(migrationops.MigrationOps, '_migrate_disks_from_source')
    def test_finish_migration(self, mock_migrate_disks_from_source,
                              mock_import_and_setup_vm):
        mock_instance = fake_instance.fake_instance_obj(self.context)
        mock_migration = mock.MagicMock()

        self._migrationops.finish_migration(
            context=self.context, migration=mock_migration,
            instance=mock_instance, disk_info=mock.sentinel.disk_info,
            network_info=mock.sentinel.network_info,
            image_meta=mock.sentinel.image_meta, resize_instance=False,
            block_device_info=mock.sentinel.block_device_info)

        mock_migrate_disks_from_source.assert_called_once_with(
            mock_migration, mock_instance, mock.sentinel.disk_info)
        mock_import_and_setup_vm.assert_called_once_with(
            self.context, mock_instance,
            mock_migrate_disks_from_source.return_value,
            mock.sentinel.image_meta, mock.sentinel.block_device_info, True)
        self._vmops.power_on.assert_called_once_with(
            mock_instance, network_info=mock.sentinel.network_info)

    @mock.patch.object(migrationops.MigrationOps, '_check_ephemeral_disks')
    @mock.patch.object(migrationops.MigrationOps, '_check_and_update_disks')
    @mock.patch.object(migrationops.MigrationOps, '_update_disk_image_paths')
    @mock.patch.object(migrationops.MigrationOps, '_import_vm')
    def test_import_and_setup_vm(self, mock_import_vm,
                                 mock_update_disk_image_paths,
                                 mock_check_and_update_disks,
                                 mock_check_eph_disks):
        block_device_info = {'ephemerals': mock.sentinel.ephemerals}
        mock_instance = fake_instance.fake_instance_obj(self.context)

        self._migrationops._import_and_setup_vm(
            self.context, mock_instance, mock.sentinel.instance_dir,
            mock.sentinel.image_meta, block_device_info,
            resize_instance=mock.sentinel.resize_instance)

        get_image_vm_gen = self._vmops.get_image_vm_generation
        get_image_vm_gen.assert_called_once_with(mock_instance.uuid,
                                                 mock.sentinel.image_meta)
        mock_import_vm.assert_called_once_with(mock.sentinel.instance_dir)
        self._migrationops._vmops.update_vm_resources.assert_called_once_with(
            mock_instance, get_image_vm_gen.return_value,
            mock.sentinel.image_meta, mock.sentinel.instance_dir,
            mock.sentinel.resize_instance)
        self._migrationops._volumeops.connect_volumes.assert_called_once_with(
            block_device_info)
        mock_update_disk_image_paths.assert_called_once_with(
            mock_instance, mock.sentinel.instance_dir)
        mock_check_and_update_disks.assert_called_once_with(
            self.context, mock_instance, get_image_vm_gen.return_value,
            mock.sentinel.image_meta, block_device_info,
            resize_instance=mock.sentinel.resize_instance)
        self._volumeops.fix_instance_volume_disk_paths.assert_called_once_with(
            mock_instance.name, block_device_info)
        self._migrationops._migrationutils.realize_vm.assert_called_once_with(
            mock_instance.name)
        mock_check_eph_disks.assert_called_once_with(
            mock_instance, mock.sentinel.ephemerals,
            mock.sentinel.resize_instance)
        self._migrationops._vmops.configure_remotefx.assert_called_once_with(
            mock_instance, get_image_vm_gen.return_value,
            mock.sentinel.resize_instance)
        self._vmops.configure_instance_metrics.assert_called_once_with(
            mock_instance.name)

    def test_import_vm(self):
        self._migrationops._import_vm(mock.sentinel.instance_dir)

        self._pathutils.get_instance_snapshot_dir.assert_called_once_with(
            instance_dir=mock.sentinel.instance_dir)
        self._pathutils.get_vm_config_file.assert_called_once_with(
            self._migrationops._pathutils.get_export_dir.return_value)
        mock_import_vm_definition = (
            self._migrationops._migrationutils.import_vm_definition)
        mock_import_vm_definition.assert_called_once_with(
            self._pathutils.get_vm_config_file.return_value,
            self._pathutils.get_instance_snapshot_dir.return_value)
        self._migrationops._pathutils.get_export_dir.assert_has_calls([
            mock.call(instance_dir=mock.sentinel.instance_dir),
            mock.call(instance_dir=mock.sentinel.instance_dir,
                      remove_dir=True)])

    @mock.patch('os.path.exists')
    def test_update_disk_image_paths(self, mock_exists):
        mock_instance = fake_instance.fake_instance_obj(self.context)
        inst_dir = "instances"
        expected_inst_dir = "expected_instances"
        config_drive_iso = os.path.join(inst_dir, 'configdrive.iso')
        expected_config_drive_iso = os.path.join(expected_inst_dir,
                                                 'configdrive.iso')
        ephemeral_disk = os.path.join(inst_dir, 'eph1.vhdx')
        expected_ephemeral_disk = os.path.join(expected_inst_dir, 'eph1.vhdx')
        other_disk = '//some/path/to/vol-UUID.vhdx'
        disk_files = [config_drive_iso, ephemeral_disk, other_disk]

        self._vmutils.get_vm_storage_paths.return_value = (
            disk_files, mock.sentinel.volume_drives)
        mock_exists.return_value = True

        self._migrationops._update_disk_image_paths(mock_instance,
                                                    expected_inst_dir)

        self._vmutils.get_vm_storage_paths.assert_called_once_with(
            mock_instance.name)
        expected_calls = [
            mock.call(config_drive_iso, expected_config_drive_iso,
                      is_physical=False),
            mock.call(ephemeral_disk, expected_ephemeral_disk,
                      is_physical=False)]
        self._vmutils.update_vm_disk_path.assert_has_calls(expected_calls)

    @mock.patch('os.path.exists')
    def test_update_disk_image_paths_exception(self, mock_exists):
        mock_instance = fake_instance.fake_instance_obj(self.context)
        inst_dir = "instances"
        disk_files = [os.path.join(inst_dir, "root.vhdx")]

        self._vmutils.get_vm_storage_paths.return_value = (
            disk_files, mock.sentinel.volume_drives)
        self._pathutils.get_instance_dir.return_value = inst_dir
        mock_exists.return_value = False

        self.assertRaises(exception.DiskNotFound,
                          self._migrationops._update_disk_image_paths,
                          mock_instance, inst_dir)

        self._vmutils.get_vm_storage_paths.assert_called_once_with(
            mock_instance.name)
        self.assertFalse(self._vmutils.update_vm_disk_path.called)

    @ddt.data(constants.DISK, mock.sentinel.root_type)
    @mock.patch.object(migrationops.MigrationOps, '_check_base_disk')
    @mock.patch.object(migrationops.MigrationOps, '_check_resize_vhd')
    def test_check_and_update_disks(self, root_type,
                                    mock_check_resize_vhd,
                                    mock_check_base_disk):
        mock_instance = fake_instance.fake_instance_obj(self.context)
        mock_instance.flavor.root_gb = 1
        root_device = {'type': root_type}
        block_device_info = {'root_disk': root_device,
                             'ephemerals': mock.sentinel.ephemerals}
        expected_check_resize = []
        expected_get_info = []

        self._migrationops._check_and_update_disks(
            self.context, mock_instance, mock.sentinel.vm_gen,
            mock.sentinel.image_meta, block_device_info, resize_instance=True)

        mock_bdi = self._block_dev_man.validate_and_update_bdi
        mock_bdi.assert_called_once_with(
            mock_instance, mock.sentinel.image_meta, mock.sentinel.vm_gen,
            block_device_info)

        if root_device['type'] == constants.DISK:
            root_device_path = (
                self._pathutils.lookup_root_vhd_path.return_value)
            self._pathutils.lookup_root_vhd_path.assert_called_once_with(
                mock_instance.name)
            expected_get_info = [mock.call(root_device_path)]

            mock_vhd_info = self._vhdutils.get_vhd_info.return_value
            mock_vhd_info.get.assert_called_once_with("ParentPath")
            mock_check_base_disk.assert_called_once_with(
                self.context, mock_instance, root_device_path,
                mock_vhd_info.get.return_value)
            expected_check_resize.append(
                mock.call(root_device_path, mock_vhd_info,
                          mock_instance.flavor.root_gb * units.Gi))
        else:
            self.assertFalse(self._pathutils.lookup_root_vhd_path.called)

        mock_check_resize_vhd.assert_has_calls(expected_check_resize)
        self._vhdutils.get_vhd_info.assert_has_calls(
            expected_get_info)

    def test_check_and_update_disks_not_found(self):
        mock_instance = fake_instance.fake_instance_obj(self.context)
        root_device = {'type': constants.DISK}
        block_device_info = {'root_disk': root_device}

        self._pathutils.lookup_root_vhd_path.return_value = None

        self.assertRaises(exception.DiskNotFound,
                          self._migrationops._check_and_update_disks,
                          self.context, mock_instance, mock.sentinel.vm_gen,
                          mock.sentinel.image_meta, block_device_info,
                          resize_instance=True)

        self._pathutils.get_instance_dir.assert_called_once_with(
            mock_instance.name)

    @mock.patch.object(migrationops.MigrationOps, '_check_resize_vhd')
    @mock.patch.object(migrationops.LOG, 'warning')
    def test_check_ephemeral_disks_multiple_eph_warn(self, mock_warn,
                                                     mock_check_resize_vhd):
        mock_instance = fake_instance.fake_instance_obj(self.context)
        mock_instance.ephemeral_gb = 3
        mock_ephemerals = [{'size': 1}, {'size': 1}]

        self._migrationops._check_ephemeral_disks(mock_instance,
                                                  mock_ephemerals,
                                                  True)

        mock_warn.assert_called_once_with(
            "Cannot resize multiple ephemeral disks for instance.",
            instance=mock_instance)

    def test_check_ephemeral_disks_exception(self):
        mock_instance = fake_instance.fake_instance_obj(self.context,
                                                        ephemeral_gb=1)
        mock_ephemerals = [dict(size=1)]

        lookup_eph_path = (
            self._migrationops._pathutils.lookup_ephemeral_vhd_path)
        lookup_eph_path.return_value = None

        self.assertRaises(exception.DiskNotFound,
                          self._migrationops._check_ephemeral_disks,
                          mock_instance, mock_ephemerals)

    @ddt.data({},
              {'existing_eph_path': mock.sentinel.eph_path},
              {'existing_eph_path': mock.sentinel.eph_path,
               'new_eph_size': 0},
              {'use_default_eph': True})
    @ddt.unpack
    @mock.patch.object(migrationops.MigrationOps, '_check_resize_vhd')
    def test_check_ephemeral_disks(self, mock_check_resize_vhd,
                                   existing_eph_path=None, new_eph_size=42,
                                   use_default_eph=False):
        mock_vmops = self._migrationops._vmops

        mock_instance = fake_instance.fake_instance_obj(self.context)
        mock_instance.ephemeral_gb = new_eph_size
        eph = {}
        mock_ephemerals = [eph] if not use_default_eph else []

        mock_pathutils = self._migrationops._pathutils
        lookup_eph_path = mock_pathutils.lookup_ephemeral_vhd_path
        lookup_eph_path.return_value = existing_eph_path
        mock_get_eph_vhd_path = mock_pathutils.get_ephemeral_vhd_path
        mock_get_eph_vhd_path.return_value = mock.sentinel.get_path

        mock_vhdutils = self._migrationops._vhdutils
        mock_get_vhd_format = mock_vhdutils.get_best_supported_vhd_format
        mock_get_vhd_format.return_value = mock.sentinel.vhd_format

        self._vmutils.get_free_controller_slot.return_value = (
            mock.sentinel.ctrl_slot)

        attached_eph_paths = [mock.sentinel.eph_path,
                              mock.sentinel.default_eph_path]
        mock_vmops.get_attached_ephemeral_disks.return_value = (
            attached_eph_paths)

        self._migrationops._check_ephemeral_disks(mock_instance,
                                                  mock_ephemerals,
                                                  True)

        if not use_default_eph:
            self.assertEqual(mock_instance.ephemeral_gb, eph['size'])
        if not existing_eph_path:
            mock_vmops.create_ephemeral_disk.assert_called_once_with(
                mock_instance.name, mock.ANY)
            self._vmutils.get_vm_scsi_controller.assert_called_once_with(
                mock_instance.name)
            self._vmutils.get_free_controller_slot.assert_called_once_with(
                self._vmutils.get_vm_scsi_controller.return_value)

            create_eph_args = mock_vmops.create_ephemeral_disk.call_args_list
            created_eph = create_eph_args[0][0][1]
            self.assertEqual(mock.sentinel.vhd_format, created_eph['format'])
            self.assertEqual(mock.sentinel.get_path, created_eph['path'])
            self.assertEqual(constants.CTRL_TYPE_SCSI,
                             created_eph['disk_bus'])
            self.assertEqual(mock.sentinel.ctrl_slot,
                             created_eph['ctrl_disk_addr'])
        elif new_eph_size:
            mock_check_resize_vhd.assert_called_once_with(
                existing_eph_path,
                self._migrationops._vhdutils.get_vhd_info.return_value,
                mock_instance.ephemeral_gb * units.Gi)
            self.assertEqual(existing_eph_path, eph['path'])
        else:
            self._vmutils.detach_vm_disk.assert_has_calls(
                [mock.call(mock_instance.name, eph_path,
                          is_physical=False)
                 for eph_path in attached_eph_paths],
                any_order=True)
            self._migrationops._pathutils.remove.assert_has_calls(
                [mock.call(eph_path) for eph_path in attached_eph_paths],
                any_order=True)



# Copyright 2015 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock

import ddt
from nova import utils
from os_win import constants

from compute_hyperv.nova import eventhandler
from compute_hyperv.nova import vmops
from compute_hyperv.tests.unit import test_base


@ddt.ddt
class EventHandlerTestCase(test_base.HyperVBaseTestCase):
    _FAKE_POLLING_INTERVAL = 3
    _FAKE_EVENT_CHECK_TIMEFRAME = 15

    def setUp(self):
        super(EventHandlerTestCase, self).setUp()

        self.flags(
            power_state_check_timeframe=self._FAKE_EVENT_CHECK_TIMEFRAME,
            group='hyperv')
        self.flags(
            power_state_event_polling_interval=self._FAKE_POLLING_INTERVAL,
            group='hyperv')

        self._event_handler = eventhandler.InstanceEventHandler()

    @ddt.data(True, False)
    @mock.patch.object(vmops.VMOps, 'get_instance_uuid')
    @mock.patch.object(eventhandler.InstanceEventHandler, '_emit_event')
    def test_handle_event(self, missing_uuid, mock_emit_event, mock_get_uuid):
        mock_get_uuid.return_value = (
            mock.sentinel.instance_uuid if not missing_uuid else None)
        self._event_handler._vmutils.get_vm_power_state.return_value = (
            mock.sentinel.power_state)

        self._event_handler._handle_event(mock.sentinel.instance_name,
                                          mock.sentinel.power_state)

        if not missing_uuid:
            mock_emit_event.assert_called_once_with(
                mock.sentinel.instance_name,
                mock.sentinel.instance_uuid,
                mock.sentinel.power_state)
        else:
            self.assertFalse(mock_emit_event.called)

    @mock.patch.object(eventhandler.InstanceEventHandler, '_get_virt_event')
    @mock.patch.object(utils, 'spawn_n',
                       lambda f, *args, **kwargs: f(*args, **kwargs))
    def test_emit_event(self, mock_get_event):
        state = constants.HYPERV_VM_STATE_ENABLED
        callbacks = [mock.Mock(), mock.Mock()]

        for cbk in callbacks:
            self._event_handler.add_callback(cbk)

        self._event_handler._emit_event(mock.sentinel.instance_name,
                                        mock.sentinel.instance_uuid,
                                        state)

        for cbk in callbacks:
            cbk.assert_called_once_with(mock_get_event.return_value)

    def test_get_virt_event(self):
        instance_state = constants.HYPERV_VM_STATE_ENABLED
        expected_transition = self._event_handler._TRANSITION_MAP[
            instance_state]

        virt_event = self._event_handler._get_virt_event(
            mock.sentinel.instance_uuid,
            mock.sentinel.instance_name,
            instance_state)

        self.assertEqual(mock.sentinel.instance_name, virt_event.name)
        self.assertEqual(mock.sentinel.instance_uuid, virt_event.uuid)
        self.assertEqual(expected_transition, virt_event.transition)
